from typing import Literal

from pydantic import model_validator
from typing_extensions import Self, override

from pipelex import log
from pipelex.cogt.content_generation.content_generator_dry import ContentGeneratorDry
from pipelex.cogt.content_generation.content_generator_protocol import ContentGeneratorProtocol
from pipelex.cogt.exceptions import ModelChoiceNotFoundError
from pipelex.cogt.extract.extract_input import ExtractInput
from pipelex.cogt.extract.extract_job_components import ExtractJobConfig, ExtractJobParams
from pipelex.cogt.extract.extract_setting import ExtractModelChoice, ExtractSetting
from pipelex.cogt.models.model_deck_check import check_extract_choice_with_deck
from pipelex.core.concepts.native.concept_native import NativeConceptCode
from pipelex.core.memory.working_memory import WorkingMemory
from pipelex.core.pipes.exceptions import PipeValidationError, PipeValidationErrorType
from pipelex.core.pipes.inputs.input_requirements import InputRequirements
from pipelex.core.pipes.pipe_output import PipeOutput
from pipelex.core.stuffs.image_content import ImageContent
from pipelex.core.stuffs.list_content import ListContent
from pipelex.core.stuffs.page_content import PageContent
from pipelex.core.stuffs.stuff_factory import StuffFactory
from pipelex.core.stuffs.text_and_images_content import TextAndImagesContent
from pipelex.core.stuffs.text_content import TextContent
from pipelex.hub import (
    get_content_generator,
    get_model_deck,
    get_native_concept,
)
from pipelex.pipe_operators.pipe_operator import PipeOperator
from pipelex.pipe_run.exceptions import PipeRunError
from pipelex.pipe_run.pipe_run_mode import PipeRunMode
from pipelex.pipe_run.pipe_run_params import PipeRunParams
from pipelex.pipeline.job_metadata import JobMetadata
from pipelex.tools.pdf.pypdfium2_renderer import pypdfium2_renderer


class PipeExtractOutput(PipeOutput):
    pass


class PipeExtract(PipeOperator[PipeExtractOutput]):
    type: Literal["PipeExtract"] = "PipeExtract"
    extract_choice: ExtractModelChoice | None
    should_caption_images: bool
    should_include_images: bool
    should_include_page_views: bool
    page_views_dpi: int

    image_stuff_name: str | None = None
    pdf_stuff_name: str | None = None

    @override
    def needed_inputs(self, visited_pipes: set[str] | None = None) -> InputRequirements:
        return self.inputs

    @override
    def required_variables(self) -> set[str]:
        return set(self.inputs.required_names)

    @model_validator(mode="after")
    def validate_fields(self) -> Self:
        if self.image_stuff_name is None and self.pdf_stuff_name is None:
            msg = "For PipeExtract you must provide either a pdf or an image or a concept that refines one of them"
            raise ValueError(msg)
        return self

    @override
    def validate_inputs_static(self):
        if self.extract_choice:
            try:
                check_extract_choice_with_deck(extract_choice=self.extract_choice)
            except ModelChoiceNotFoundError as exc:
                msg = f"Extract choice '{self.extract_choice}' was not found in the model deck"
                raise ValueError(msg) from exc

    @override
    def validate_inputs_with_library(self):
        pass

    @override
    def validate_output_static(self):
        pass

    @override
    def validate_output_with_library(self):
        if self.output != get_native_concept(native_concept=NativeConceptCode.PAGE):
            msg = f"PipeExtract output should be a Page concept, but is {self.output.concept_string}"
            raise PipeValidationError(
                message=msg,
                error_type=PipeValidationErrorType.INADEQUATE_OUTPUT_CONCEPT,
                domain=self.domain,
                pipe_code=self.code,
                provided_concept_code=self.output.concept_string,
                required_concept_codes=[NativeConceptCode.PAGE.concept_string],
            )

    @override
    async def _run_operator_pipe(
        self,
        job_metadata: JobMetadata,
        working_memory: WorkingMemory,
        pipe_run_params: PipeRunParams,
        output_name: str | None = None,
        content_generator: ContentGeneratorProtocol | None = None,
    ) -> PipeExtractOutput:
        content_generator = content_generator or get_content_generator()

        image_uri: str | None = None
        pdf_uri: str | None = None
        if self.image_stuff_name:
            image_stuff = working_memory.get_stuff_as_image(name=self.image_stuff_name)
            image_uri = image_stuff.url
        elif self.pdf_stuff_name:
            pdf_stuff = working_memory.get_stuff_as_pdf(name=self.pdf_stuff_name)
            pdf_uri = pdf_stuff.url

        extract_choice: ExtractModelChoice = self.extract_choice or get_model_deck().extract_choice_default
        extract_setting: ExtractSetting = get_model_deck().get_extract_setting(extract_choice=extract_choice)

        extract_job_params = ExtractJobParams(
            should_include_images=self.should_include_images,
            should_caption_images=self.should_caption_images,
            should_include_page_views=self.should_include_page_views,
            page_views_dpi=self.page_views_dpi,
            max_nb_images=extract_setting.max_nb_images,
            image_min_size=extract_setting.image_min_size,
        )
        extract_input = ExtractInput(
            image_uri=image_uri,
            pdf_uri=pdf_uri,
        )
        extract_output = await content_generator.make_extract_pages(
            extract_input=extract_input,
            extract_handle=extract_setting.model,
            job_metadata=job_metadata,
            extract_job_params=extract_job_params,
            extract_job_config=ExtractJobConfig(),
        )

        # Build the output stuff, which is a list of page contents
        page_view_contents: list[ImageContent] = []
        if self.should_include_page_views:
            log.verbose(f"should_include_page_views: {self.should_include_page_views}, pdf_uri: {pdf_uri}, image_uri: {image_uri}")
            if pdf_uri:
                page_view_contents.extend(
                    ImageContent.make_from_extracted_image(extracted_image=page.page_view) for page in extract_output.pages.values() if page.page_view
                )
                log.verbose(f"page_view_contents: {page_view_contents}")
                needs_to_generate_page_views: bool
                if len(page_view_contents) == 0:
                    log.verbose("No page views found in the OCR output")
                    needs_to_generate_page_views = True
                elif len(page_view_contents) < len(extract_output.pages):
                    log.warning(f"Only {len(page_view_contents)} page found in the OCR output, but {len(extract_output.pages)} pages")
                    needs_to_generate_page_views = True
                else:
                    log.verbose("All page views found in the OCR output")
                    needs_to_generate_page_views = False

                if needs_to_generate_page_views:
                    page_views = await pypdfium2_renderer.render_pdf_pages_from_uri(pdf_uri=pdf_uri, dpi=self.page_views_dpi)
                    page_view_contents = [ImageContent.make_from_image(image=img) for img in page_views]
            elif image_uri:
                page_view_contents = [ImageContent(url=image_uri)]

        page_contents: list[PageContent] = []
        for page_index, page in extract_output.pages.items():
            images = [ImageContent.make_from_extracted_image(extracted_image=img) for img in page.extracted_images]
            log.verbose(f"images: {images}, page_view_contents: {page_view_contents}, index: {page_index}")
            page_view = page_view_contents[page_index - 1] if self.should_include_page_views else None
            page_contents.append(
                PageContent(
                    text_and_images=TextAndImagesContent(
                        text=TextContent(text=page.text) if page.text else None,
                        images=images,
                    ),
                    page_view=page_view,
                ),
            )

        content: ListContent[PageContent] = ListContent(items=page_contents)

        output_stuff = StuffFactory.make_stuff(
            name=output_name,
            concept=self.output,
            content=content,
        )

        working_memory.set_new_main_stuff(
            stuff=output_stuff,
            name=output_name,
        )

        return PipeExtractOutput(
            working_memory=working_memory,
            pipeline_run_id=job_metadata.pipeline_run_id,
        )

    @override
    async def _dry_run_operator_pipe(
        self,
        job_metadata: JobMetadata,
        working_memory: WorkingMemory,
        pipe_run_params: PipeRunParams,
        output_name: str | None = None,
    ) -> PipeExtractOutput:
        log.verbose(f"PipeExtract: dry run operator pipe: {self.code}")
        if pipe_run_params.run_mode != PipeRunMode.DRY:
            msg = f"Running pipe '{self.code}' (PipeExtract) _dry_run_operator_pipe() in non-dry mode is not allowed."
            raise PipeRunError(message=msg, run_mode=pipe_run_params.run_mode)

        content_generator_dry = ContentGeneratorDry()
        return await self._run_operator_pipe(
            job_metadata=job_metadata,
            working_memory=working_memory,
            pipe_run_params=pipe_run_params,
            output_name=output_name,
            content_generator=content_generator_dry,
        )



from importlib.abc import Traversable
from importlib.resources import files
from pathlib import Path

from pipelex import log
from pipelex.config import get_config
from pipelex.core.interpreter.interpreter import PipelexInterpreter
from pipelex.tools.misc.file_utils import find_files_in_dir


def get_pipelex_plx_files_from_package() -> list[Path]:
    """Get all PLX files from the pipelex package using importlib.resources.

    This works reliably whether pipelex is installed as a wheel, from source,
    or as a relative path import.

    Returns:
        List of Path objects to PLX files in pipelex package
    """
    plx_files: list[Path] = []
    pipelex_package = files("pipelex")

    def _find_plx_in_traversable(traversable: Traversable, collected: list[Path]) -> None:
        """Recursively find .plx files in a Traversable."""
        excluded_dirs = get_config().pipelex.scan_config.excluded_dirs
        try:
            if not traversable.is_dir():
                return

            for child in traversable.iterdir():
                if child.is_file() and child.name.endswith(".plx"):
                    # Convert to path string for validation
                    plx_path_str = str(child)
                    if PipelexInterpreter.is_pipelex_file(Path(plx_path_str)):
                        collected.append(Path(plx_path_str))
                        log.verbose(f"Found pipelex package PLX file: {plx_path_str}")
                elif child.is_dir():
                    # Skip excluded directories
                    if child.name not in excluded_dirs:
                        _find_plx_in_traversable(child, collected)
        except (PermissionError, OSError) as exc:
            log.warning(f"Could not access {traversable}: {exc}")

    _find_plx_in_traversable(pipelex_package, plx_files)
    log.verbose(f"Found {len(plx_files)} PLX files in pipelex package")
    return plx_files


def get_pipelex_package_dir_for_imports() -> Path | None:
    """Get the pipelex package directory as a Path for importing Python modules.

    Returns:
        Path to the pipelex package directory, or None if not accessible as filesystem
    """
    pipelex_package = files("pipelex")
    try:
        # Try to convert to Path (works for filesystem paths)
        pkg_path = Path(str(pipelex_package))
        if pkg_path.exists() and pkg_path.is_dir():
            return pkg_path
    except (TypeError, ValueError, OSError) as exc:
        log.warning(f"Could not convert importlib.resources Traversable to filesystem Path: {exc}")
    return None


def get_pipelex_plx_files_from_dirs(dirs: set[Path]) -> list[Path]:
    """Get all valid Pipelex PLX files from the given directories."""
    all_plx_paths: list[Path] = []

    for dir_path in dirs:
        if not dir_path.exists():
            log.debug(f"Directory does not exist, skipping: {dir_path}")
            continue

        # Find all .plx files in the directory, excluding problematic directories
        plx_files = find_files_in_dir(
            dir_path=str(dir_path),
            pattern="*.plx",
            excluded_dirs=list(get_config().pipelex.scan_config.excluded_dirs),
        )

        # Filter to only include valid Pipelex files
        for plx_file in plx_files:
            if PipelexInterpreter.is_pipelex_file(plx_file):
                all_plx_paths.append(plx_file)
            else:
                log.debug(f"Skipping non-Pipelex PLX file: {plx_file}")

    return all_plx_paths



from pipelex.base_exceptions import PipelexError
from pipelex.core.bundles.exceptions import PipelexBundleBlueprintValidationErrorData
from pipelex.tools.misc.toml_utils import TomlError


class PipelexInterpreterError(PipelexError):
    """Raised when PipelexInterpreter fails."""

    def __init__(
        self,
        message: str,
        validation_errors: list[PipelexBundleBlueprintValidationErrorData] | None = None,
    ):
        self.validation_errors = validation_errors or []
        super().__init__(message)


class PLXDecodeError(TomlError):
    """Raised when PLX decoding fails."""



import re
from typing import Any

from pydantic import Field, field_validator
from rich.console import Group
from rich.table import Table
from rich.text import Text
from typing_extensions import override

from pipelex import log
from pipelex.core.concepts.validation import validate_concept_string_or_code
from pipelex.core.pipes.pipe_blueprint import PipeBlueprint, PipeCategory, PipeType
from pipelex.core.pipes.variable_multiplicity import MUTLIPLICITY_PATTERN, parse_concept_with_multiplicity
from pipelex.core.stuffs.structured_content import StructuredContent
from pipelex.tools.misc.pretty import PrettyPrintable
from pipelex.tools.misc.string_utils import is_snake_case, normalize_to_ascii


class PipeSpec(StructuredContent):
    """Spec defining a pipe: an executable component with a clear contract defined by its inputs and output.

    Pipes are the building blocks of a Pipelex pipeline. There are two categories:
    - Controllers: Manage execution flow (PipeSequence, PipeParallel, PipeCondition, PipeBatch)
    - Operators: Perform specific tasks (PipeLLM, PipeImgGen, PipeExtract, PipeFunc, PipeCompose)

    Multiplicity Notation:
        Both inputs and outputs use bracket notation to specify item counts:
        - No brackets: single item (default)
        - []: variable-length list (e.g., "Text[]")
        - [N]: exactly N items (e.g., "Image[3]" for 3 images)

    Examples:
        inputs = {"document": "PDF", "queries": "Text[]"}  # single PDF, multiple texts
        output = "Article[]"  # produces a list of articles
        output = "Image[5]"  # produces exactly 5 images
    """

    pipe_code: str = Field(description="Unique pipe identifier. Must be snake_case.")
    type: Any = Field(
        description=(
            f"Pipe type. Validated at runtime, must be one of: {PipeType}. Examples: PipeLLM, PipeImgGen, PipeExtract, PipeSequence, PipeParallel."
        )
    )
    pipe_category: Any = Field(
        description=(f"Pipe category. Validated at runtime, must be one of: {PipeCategory}. Either 'PipeController' or 'PipeOperator'.")
    )
    description: str | None = Field(description="Natural language description of the pipe's purpose and functionality.")
    inputs: dict[str, str] = Field(
        description=(
            "Input specifications mapping variable names to concept codes with optional multiplicity. "
            "Keys: input names in snake_case. "
            "Values: ConceptCodes in PascalCase with optional brackets. "
            "Examples: 'Text' (single), 'Text[]' (variable list), 'Image[2]' (exactly 2 images), 'domain.Concept[]' (domain-qualified list)."
        )
    )
    output: str = Field(
        description=(
            "Output concept code in PascalCase with optional multiplicity brackets. "
            "Examples: 'Text' (single text), 'Article[]' (list of articles), 'Image[3]' (exactly 3 images). "
            "IMPORTANT: Always use PascalCase for the concept name."
        )
    )

    @field_validator("pipe_code", mode="before")
    @classmethod
    def validate_pipe_code(cls, value: str) -> str:
        return cls.validate_pipe_code_syntax(value)

    @field_validator("type", mode="after")
    @classmethod
    def validate_pipe_type(cls, value: Any) -> Any:
        if value not in PipeType.value_list():
            msg = f"Invalid pipe type '{value}'. Must be one of: {PipeType.value_list()}"
            raise ValueError(msg)
        return value

    @field_validator("output", mode="after")
    @classmethod
    def validate_output(cls, output: str) -> str:
        # Extract concept without multiplicity for validation
        parse_result = parse_concept_with_multiplicity(output)
        validate_concept_string_or_code(concept_string_or_code=parse_result.concept)
        return output  # Return original with brackets intact

    @field_validator("inputs", mode="after")
    @classmethod
    def validate_inputs(cls, inputs: dict[str, str] | None) -> dict[str, str] | None:
        if inputs is None:
            return None

        for input_name, concept_spec in inputs.items():
            if not is_snake_case(input_name):
                msg = f"Invalid input name syntax '{input_name}'. Must be in snake_case."
                raise ValueError(msg)

            # Validate the concept spec format with optional multiplicity brackets
            # Pattern allows: ConceptName, domain.ConceptName, ConceptName[], ConceptName[N]
            match = re.match(MUTLIPLICITY_PATTERN, concept_spec)
            if not match:
                msg = (
                    f"Invalid input syntax for '{input_name}': '{concept_spec}'. "
                    f"Expected format: 'ConceptName', 'ConceptName[]', or 'ConceptName[N]' where N is an integer."
                )
                raise ValueError(msg)

            # Extract the concept part (without multiplicity) and validate it
            concept_string_or_code = match.group(1)
            validate_concept_string_or_code(concept_string_or_code=concept_string_or_code)

        return inputs

    @classmethod
    def validate_pipe_code_syntax(cls, pipe_code: str) -> str:
        # First, normalize Unicode to ASCII to prevent homograph attacks
        normalized_pipe_code = normalize_to_ascii(pipe_code)

        if normalized_pipe_code != pipe_code:
            log.warning(f"Pipe code '{pipe_code}' contained non-ASCII characters, normalized to '{normalized_pipe_code}'")

        if not is_snake_case(normalized_pipe_code):
            msg = f"Invalid pipe code syntax '{normalized_pipe_code}'. Must be in snake_case."
            raise ValueError(msg)
        return normalized_pipe_code

    def to_blueprint(self) -> PipeBlueprint:
        return PipeBlueprint(
            description=self.description,
            inputs=self.inputs if self.inputs else None,
            output=self.output,
            type=self.type,
            pipe_category=self.pipe_category,
        )

    @override
    def rendered_pretty(self, title: str | None = None, depth: int = 0) -> PrettyPrintable:
        pipe_group = Group()
        if title:
            pipe_group.renderables.append(Text(title, style="bold"))
        pipe_group.renderables.append(Text.from_markup(f"Pipe: [bold red]{self.pipe_code}[/bold red]\n"))
        pipe_group.renderables.append(Text.from_markup(f"Type: [bold magenta]{self.type}[/bold magenta] ({self.pipe_category})\n"))
        if self.description:
            pipe_group.renderables.append(Text.from_markup(f"Description: [yellow italic]{self.description}[/yellow italic]\n"))

        # Create inputs section
        if not self.inputs:
            pipe_group.renderables.append(Text.from_markup("\nNo inputs"))
        elif len(self.inputs) == 1:
            # Single input: display as a simple line of text
            input_name, concept_spec = next(iter(self.inputs.items()))
            pipe_group.renderables.append(Text.from_markup(f"\nInput: [cyan]{input_name}[/cyan] ([bold green]{concept_spec}[/bold green])"))
        else:
            # Multiple inputs: display as a table
            inputs_table = Table(
                title="Inputs:",
                title_justify="left",
                title_style="not italic",
                show_header=False,
                show_edge=True,
                show_lines=True,
                border_style="dim",
            )
            inputs_table.add_column("Variable Name", style="cyan")
            inputs_table.add_column("Concept", style="bold green")
            for input_name, concept_spec in self.inputs.items():
                inputs_table.add_row(input_name, concept_spec)
            pipe_group.renderables.append(inputs_table)

        # Show output
        pipe_group.renderables.append(Text.from_markup(f"\nOutput: [bold green]{self.output}[/bold green]"))

        return pipe_group



from typing import TYPE_CHECKING, Any, Generic, Protocol, TypeVar

from kajson.exceptions import ClassRegistryInheritanceError, ClassRegistryNotFoundError
from kajson.kajson_manager import KajsonManager
from typing_extensions import runtime_checkable

from pipelex.core.concepts.concept import Concept
from pipelex.core.concepts.concept_factory import ConceptFactory
from pipelex.core.concepts.exceptions import ConceptFactoryError
from pipelex.core.concepts.helpers import strip_multiplicity_from_concept_string_or_code
from pipelex.core.concepts.native.concept_native import NativeConceptCode
from pipelex.core.pipes.exceptions import PipeFactoryError, PipeVariableMultiplicityError
from pipelex.core.pipes.inputs.input_requirements import InputRequirements
from pipelex.core.pipes.inputs.input_requirements_factory import InputRequirementsFactory
from pipelex.core.pipes.pipe_blueprint import PipeBlueprint, PipeType
from pipelex.core.pipes.variable_multiplicity import parse_concept_with_multiplicity
from pipelex.hub import get_required_concept

if TYPE_CHECKING:
    from pipelex.core.pipes.pipe_abstract import PipeAbstract

PipeBlueprintType = TypeVar("PipeBlueprintType", bound="PipeBlueprint", contravariant=True)
PipeAbstractType = TypeVar("PipeAbstractType", bound="PipeAbstract", covariant=True)


@runtime_checkable
class PipeFactoryProtocol(Protocol[PipeBlueprintType, PipeAbstractType]):
    @classmethod
    def make(
        cls,
        pipe_category: Any,
        pipe_type: str,
        pipe_code: str,
        domain_code: str,
        description: str | None,
        inputs: InputRequirements,
        output: Concept,
        blueprint: PipeBlueprintType,
    ) -> PipeAbstractType: ...


class PipeFactory(Generic[PipeAbstractType]):
    @classmethod
    def make_from_blueprint(
        cls,
        domain_code: str,
        pipe_code: str,
        blueprint: PipeBlueprint,
        concept_codes_from_the_same_domain: list[str] | None = None,
    ) -> PipeAbstractType:
        if concept_codes_from_the_same_domain is None:
            concept_codes_from_the_same_domain = []

        # TODO: This test should move to the PipelexBlueprint validation.
        # Validate that the specified concepts are declared in the bundle, or are natives concepts.
        if blueprint.inputs is not None:
            for input_name, input_concept_string_or_code in blueprint.inputs.items():
                stripped_input_concept_string_or_code = strip_multiplicity_from_concept_string_or_code(
                    concept_string_or_code=input_concept_string_or_code
                )
                if "." not in stripped_input_concept_string_or_code:
                    if (
                        not NativeConceptCode.is_native_concept_string_or_code(concept_string_or_code=stripped_input_concept_string_or_code)
                        and stripped_input_concept_string_or_code not in concept_codes_from_the_same_domain
                    ):
                        msg = (
                            f"Input stuff '{input_name}' with concept '{stripped_input_concept_string_or_code}' "
                            f"in pipe '{pipe_code}' (domain '{domain_code}') is invalid. "
                            f"The concept must be either native, declared in domain '{domain_code}', or fully qualified with a domain prefix. "
                            f"Declared concepts are: '{concept_codes_from_the_same_domain}'"
                        )
                        raise PipeFactoryError(msg)

        if "." not in blueprint.output:
            stripped_output_concept_string_or_code = strip_multiplicity_from_concept_string_or_code(concept_string_or_code=blueprint.output)
            if (
                not NativeConceptCode.is_native_concept_string_or_code(concept_string_or_code=stripped_output_concept_string_or_code)
                and stripped_output_concept_string_or_code not in concept_codes_from_the_same_domain
            ):
                msg = (
                    f"Output concept '{stripped_output_concept_string_or_code}' in pipe '{pipe_code}' (domain '{domain_code}') is invalid. "
                    f"The concept must be either native, declared in domain '{domain_code}', or fully qualified with a domain prefix. "
                    f"Declared concepts are: '{concept_codes_from_the_same_domain}'"
                )
                raise PipeFactoryError(msg)

        # Parse common attributes
        parsed_output = cls._parse_output_concept_string(domain=domain_code, pipe_code=pipe_code, output_string=blueprint.output)
        parsed_inputs = InputRequirementsFactory.make_from_blueprint(
            domain=domain_code,
            blueprint=blueprint.inputs or {},
        )

        pipe_type = PipeType(blueprint.type)
        pipe_category = pipe_type.category

        # The factory class name for that specific type of Pipe is the pipe class name with "Factory" suffix
        factory_class_name = f"{pipe_type.value}Factory"
        try:
            pipe_factory: type[PipeFactoryProtocol[Any, Any]] = KajsonManager.get_class_registry().get_required_subclass(
                name=factory_class_name,
                base_class=PipeFactoryProtocol,
            )
        except ClassRegistryNotFoundError as factory_not_found_error:
            msg = f"Pipe '{pipe_code}' couldn't be created: factory '{factory_class_name}' not found: {factory_not_found_error}"
            raise PipeFactoryError(msg) from factory_not_found_error
        except ClassRegistryInheritanceError as factory_inheritance_error:
            msg = f"Pipe '{pipe_code}' couldn't be created: factory '{factory_class_name}' is not a subclass of {type(PipeFactoryProtocol)}."
            raise PipeFactoryError(msg) from factory_inheritance_error

        pipe: PipeAbstractType = pipe_factory.make(
            pipe_category=pipe_category,
            pipe_type=blueprint.type,
            pipe_code=pipe_code,
            domain_code=domain_code,
            description=blueprint.description,
            inputs=parsed_inputs,
            output=parsed_output,
            blueprint=blueprint,
        )
        return pipe

    @classmethod
    def _parse_output_concept_string(
        cls,
        domain: str,
        pipe_code: str,
        output_string: str,
    ) -> Concept:
        """Parse the output concept string and return the Concept object."""
        # Parse output to strip multiplicity brackets
        try:
            output_parse_result = parse_concept_with_multiplicity(output_string)
        except PipeVariableMultiplicityError as exc:
            msg = f"Error parsing concept with multiplicity for pipe '{pipe_code}': {exc}"
            raise PipeFactoryError(msg) from exc

        # Get output concept
        try:
            output_domain_and_code = ConceptFactory.make_domain_and_concept_code_from_concept_string_or_code(
                domain=domain,
                concept_string_or_code=output_parse_result.concept,
            )
        except ConceptFactoryError as exc:
            msg = f"Error making domain and concept code for pipe '{pipe_code}': {exc}"
            raise PipeFactoryError(msg) from exc

        return get_required_concept(
            concept_string=ConceptFactory.make_concept_string_with_domain(
                domain=output_domain_and_code.domain,
                concept_code=output_domain_and_code.concept_code,
            ),
        )



from pydantic import BaseModel
from typing_extensions import override

from pipelex import log
from pipelex.cogt.exceptions import ImgGenPromptError
from pipelex.system.runtime import ProblemReaction, runtime_manager
from pipelex.tools.misc.json_utils import json_str


class ImgGenPrompt(BaseModel):
    positive_text: str

    def validate_before_execution(self):
        reaction = runtime_manager.problem_reactions.job
        match reaction:
            case ProblemReaction.NONE:
                pass
            case ProblemReaction.RAISE:
                if self.positive_text == "":
                    msg = "ImgGen prompt positive_text must not be an empty string"
                    raise ImgGenPromptError(msg)
            case ProblemReaction.LOG:
                if self.positive_text == "":
                    log.warning("ImgGen prompt positive_text should not be an empty string")

    @override
    def __str__(self) -> str:
        return json_str(self, title="img_gen_prompt", is_spaced=True)



import inspect
from typing import Any, cast, get_args, get_origin

from kajson.kajson_manager import KajsonManager
from pydantic import BaseModel, ConfigDict, field_validator

from pipelex import log
from pipelex.base_exceptions import PipelexUnexpectedError
from pipelex.core.concepts.exceptions import ConceptCodeError, ConceptStringError, ConceptValueError
from pipelex.core.concepts.native.concept_native import NativeConceptCode
from pipelex.core.concepts.validation import validate_concept_code, validate_concept_string
from pipelex.core.domains.exceptions import DomainCodeError
from pipelex.core.domains.validation import validate_domain_code
from pipelex.core.stuffs.image_field_search import search_for_nested_image_fields
from pipelex.core.stuffs.stuff_content import StuffContent
from pipelex.tools.misc.string_utils import pascal_case_to_sentence
from pipelex.tools.typing.class_utils import are_classes_equivalent, has_compatible_field
from pipelex.types import StrEnum


class Concept(BaseModel):
    model_config = ConfigDict(extra="forbid", strict=True)

    code: str
    domain: str
    description: str
    structure_class_name: str
    refines: str | None = None

    @field_validator("code")
    @classmethod
    def validate_code(cls, code: str) -> str:
        try:
            validate_concept_code(concept_code=code)
        except ConceptCodeError as exc:
            msg = f"Concept code '{code}' is not a valid concept code for concept '{cls.concept_string}'"
            raise ConceptValueError(msg) from exc
        return code

    @field_validator("domain")
    @classmethod
    def validate_domain(cls, domain: str) -> str:
        try:
            validate_domain_code(code=domain)
        except DomainCodeError as exc:
            msg = f"Domain code '{domain}' is not a valid domain code for concept '{cls.concept_string}'"
            raise ConceptValueError(msg) from exc
        return domain

    @field_validator("refines", mode="before")
    @classmethod
    def validate_refines(cls, refines: str | None) -> str | None:
        if refines is None:
            return None
        try:
            validate_concept_string(concept_string=refines)
        except ConceptStringError as exc:
            msg = f"Refines '{refines}' is not a valid concept string for concept '{cls.concept_string}'"
            raise ConceptValueError(msg) from exc
        return refines

    @property
    def concept_string(self) -> str:
        return f"{self.domain}.{self.code}"

    @classmethod
    def sentence_from_concept(cls, concept: "Concept") -> str:
        return pascal_case_to_sentence(name=concept.code)

    @classmethod
    def is_native_concept(cls, concept: "Concept") -> bool:
        return NativeConceptCode.is_native_concept_string_or_code(concept_string_or_code=concept.concept_string)

    @classmethod
    def are_concept_compatible(cls, concept_1: "Concept", concept_2: "Concept", strict: bool = False) -> bool:
        if NativeConceptCode.is_dynamic_concept(concept_code=concept_1.code):
            return True
        if NativeConceptCode.is_dynamic_concept(concept_code=concept_2.code):
            return True
        if concept_1.concept_string == concept_2.concept_string:
            return True
        if concept_1.structure_class_name == concept_2.structure_class_name:
            return True

        # If concept_1 refines concept_2, they are strictly compatible
        if concept_1.refines is not None and concept_1.refines == concept_2.concept_string:
            return True

        if concept_1.refines is None and concept_2.refines is None:
            concept_1_class = KajsonManager.get_class_registry().get_class(name=concept_1.structure_class_name)
            concept_2_class = KajsonManager.get_class_registry().get_class(name=concept_2.structure_class_name)

            if concept_1_class is None or concept_2_class is None:
                return False

            if strict:
                # Check if classes are equivalent (same fields, types, descriptions)
                return are_classes_equivalent(concept_1_class, concept_2_class)
            # Check if concept_1 is a subclass of concept_2
            try:
                if issubclass(concept_1_class, concept_2_class):
                    return True
            except TypeError:
                pass

            # Check if concept_1 has compatible fields with concept_2
            return has_compatible_field(concept_1_class, concept_2_class)
        return False

    @classmethod
    def is_valid_structure_class(cls, structure_class_name: str) -> bool:
        # TODO: DO NOT use the KajsonManager here. Pipelex needs to be instantiated to use the get_class_registry.
        # And when we go through KajsonManager, no error raises if pipelex is not instantiated.
        # We get_class_registry directly from KajsonManager instead of pipelex hub to avoid circular import
        if KajsonManager.get_class_registry().has_subclass(name=structure_class_name, base_class=StuffContent):
            return True
        # We get_class_registry directly from KajsonManager instead of pipelex hub to avoid circular import
        if KajsonManager.get_class_registry().has_class(name=structure_class_name):
            log.warning(f"Concept class '{structure_class_name}' is registered but it's not a subclass of StuffContent")
        return False

    def search_for_nested_image_fields_in_structure_class(self) -> list[str]:
        """Recursively search for image fields in a structure class."""
        structure_class = KajsonManager.get_class_registry().get_required_subclass(name=self.structure_class_name, base_class=StuffContent)
        if not issubclass(structure_class, StuffContent):
            msg = f"Concept class '{self.structure_class_name}' is not a subclass of StuffContent"
            raise PipelexUnexpectedError(msg)
        return search_for_nested_image_fields(content_class=structure_class)

    def get_compact_memory_example(self, var_name: str) -> dict[str, Any] | str | int:
        """Generate an example value for compact memory format based on this concept.

        Compact memory follows these conventions:
        - For native concepts that can be represented as simple values (Text, Image, PDF): returns a simple string
        - For structured concepts: returns {"concept_code": "...", "content": {...}}

        The content dict is recursively generated based on the StuffContent class structure.
        """
        # Get the structure class
        structure_class = KajsonManager.get_class_registry().get_class(name=self.structure_class_name)

        # If class not found, return placeholder
        if structure_class is None:
            return {
                "concept_code": self.concept_string,
                "content": {},  # Empty dict for unknown structures
            }

        # Verify it's a subclass of StuffContent
        if not issubclass(structure_class, StuffContent):
            return {
                "concept_code": self.concept_string,
                "content": {},  # Empty dict for invalid structures
            }

        # Generate the content based on structure
        content_example = self._generate_content_example_for_class(structure_class, var_name)

        # Check if this is actually a native concept (not just using a native structure class)
        is_native = Concept.is_native_concept(self)

        # For simple native concepts ONLY - return compact format
        if is_native and self.structure_class_name == "TextContent":
            return cast("str", content_example)  # Just a string
        elif is_native and self.structure_class_name == "ImageContent":
            # Return dict with class instantiation info
            return {
                "_class": "ImageContent",
                "url": cast("str", content_example),
            }
        elif is_native and self.structure_class_name == "PDFContent":
            # Return dict with class instantiation info
            return {
                "_class": "PDFContent",
                "url": cast("str", content_example),
            }
        elif is_native and self.structure_class_name == "NumberContent":
            return cast("int", content_example)  # Just a number

        # For refined or complex concepts, wrap with concept_code
        # For Image/PDF content, wrap in the _class format
        if self.structure_class_name == "ImageContent":
            content_wrapped = {
                "_class": "ImageContent",
                "url": cast("str", content_example),
            }
        elif self.structure_class_name == "PDFContent":
            content_wrapped = {
                "_class": "PDFContent",
                "url": cast("str", content_example),
            }
        else:
            content_wrapped = content_example

        return {
            "concept_code": self.concept_string,
            "content": content_wrapped,
        }

    @classmethod
    def _generate_content_example_for_class(cls, content_class: type[StuffContent], var_name: str) -> Any:
        """Recursively generate example content based on a StuffContent class structure.

        Args:
            content_class: The StuffContent class to generate an example for
            var_name: Variable name for generating contextual example values

        Returns:
            Example content dict or simple value
        """
        class_name = content_class.__name__

        # Handle simple native content types
        if class_name == "TextContent":
            return f"{var_name}_text"
        elif class_name in {"ImageContent", "PDFContent"}:
            return f"{var_name}_url"
        elif class_name == "NumberContent":
            return 0

        # For structured content, inspect fields and recursively generate
        # Note: model_fields includes inherited fields from parent classes
        content_dict: dict[str, Any] = {}
        for field_name, field_info in content_class.model_fields.items():
            field_type = field_info.annotation

            # Handle Optional types (e.g., TextContent | None)
            origin = get_origin(field_type)
            args = get_args(field_type)

            if origin is type(None) or (args and type(None) in args):
                # Optional field - get the non-None type
                actual_type = next((arg for arg in args if arg is not type(None)), field_type) if args else field_type
            else:
                actual_type = field_type

            # Re-check origin after unwrapping Optional
            origin = get_origin(actual_type)
            args = get_args(actual_type)

            # Handle list types
            if origin is list:
                list_item_type = args[0] if args else str
                if hasattr(list_item_type, "__name__"):
                    if list_item_type.__name__ == "ImageContent":
                        content_dict[field_name] = [f"{field_name}_url_1", f"{field_name}_url_2"]
                    elif list_item_type.__name__ == "TextContent":
                        content_dict[field_name] = [f"{field_name}_text_1", f"{field_name}_text_2"]
                    elif inspect.isclass(list_item_type) and issubclass(list_item_type, StuffContent):
                        # List of StuffContent - generate examples
                        content_dict[field_name] = [cls._generate_content_example_for_class(list_item_type, f"{field_name}_item")]
                    else:
                        content_dict[field_name] = [f"{field_name}_item_1"]
                else:
                    content_dict[field_name] = []
            # Handle dict types
            elif origin is dict:
                # Simple example with one key-value pair
                content_dict[field_name] = {f"{field_name}_key": f"{field_name}_value"}
            # Handle StrEnum types
            elif inspect.isclass(actual_type) and issubclass(actual_type, StrEnum):
                # Get first enum value
                enum_values = list(actual_type)
                content_dict[field_name] = enum_values[0].value if enum_values else f"{field_name}_enum_value"
            # Handle nested StuffContent
            elif inspect.isclass(actual_type) and issubclass(actual_type, StuffContent):
                content_dict[field_name] = cls._generate_content_example_for_class(actual_type, field_name)
            # Handle basic types
            elif actual_type is str:
                content_dict[field_name] = f"{field_name}_value"
            elif actual_type is int:
                content_dict[field_name] = 0
            elif actual_type is float:
                content_dict[field_name] = 0.0
            elif actual_type is bool:
                content_dict[field_name] = False
            else:
                # For unknown types, try to get a simple repr
                try:
                    type_name = getattr(actual_type, "__name__", str(actual_type))
                    content_dict[field_name] = f"{field_name}_value  # TODO: Fill {type_name}"
                except Exception:
                    content_dict[field_name] = f"{field_name}_value"

        return content_dict



from pathlib import Path
from typing import TYPE_CHECKING

from pipelex.client.protocol import PipelineInputs
from pipelex.core.memory.working_memory import WorkingMemory
from pipelex.core.memory.working_memory_factory import WorkingMemoryFactory
from pipelex.hub import (
    get_library_manager,
    get_pipeline_manager,
    get_report_delegate,
    get_required_pipe,
    get_telemetry_manager,
    set_current_library,
)
from pipelex.pipe_run.pipe_job import PipeJob
from pipelex.pipe_run.pipe_job_factory import PipeJobFactory
from pipelex.pipe_run.pipe_run_mode import PipeRunMode
from pipelex.pipe_run.pipe_run_params import (
    FORCE_DRY_RUN_MODE_ENV_KEY,
    VariableMultiplicity,
)
from pipelex.pipe_run.pipe_run_params_factory import PipeRunParamsFactory
from pipelex.pipeline.exceptions import PipeExecutionError
from pipelex.pipeline.job_metadata import JobMetadata
from pipelex.pipeline.validate_bundle import validate_bundle
from pipelex.system.environment import get_optional_env
from pipelex.system.telemetry.events import EventName, EventProperty

if TYPE_CHECKING:
    from pipelex.core.bundles.pipelex_bundle_blueprint import PipelexBundleBlueprint
    from pipelex.core.pipes.pipe_abstract import PipeAbstract


async def pipeline_run_setup(
    library_id: str | None = None,
    library_dirs: list[str] | None = None,
    pipe_code: str | None = None,
    plx_content: str | None = None,
    inputs: PipelineInputs | WorkingMemory | None = None,
    output_name: str | None = None,
    output_multiplicity: VariableMultiplicity | None = None,
    dynamic_output_concept_code: str | None = None,
    pipe_run_mode: PipeRunMode | None = None,
    search_domains: list[str] | None = None,
) -> tuple[PipeJob, str, str]:
    """Set up a pipeline for execution.

    This function handles all the common setup logic for both ``execute_pipeline``
    and ``start_pipeline``, including library setup, pipe loading, working memory
    initialization, and pipe job creation.

    Parameters
    ----------
    library_id:
        Unique identifier for the library instance. If not provided, defaults to the
        auto-generated ``pipeline_run_id``. Use a custom ID when you need to manage
        multiple library instances or maintain library state across executions.
    library_dirs:
        List of directory paths to load pipe definitions from. If not provided, loads
        from the current working directory (the directory from which the Python script
        is executed). Ignored when ``plx_content`` is provided.
    pipe_code:
        Code identifying the pipe to execute. Required when ``plx_content`` is not
        provided. When both ``plx_content`` and ``pipe_code`` are provided, the
        specified pipe from the PLX content will be executed (overriding any
        ``main_pipe`` defined in the content).
    plx_content:
        Complete PLX file content as a string. When provided, only this content is
        loaded into the library, creating an isolated execution environment. The pipe
        to execute is determined by ``pipe_code`` (if provided) or the ``main_pipe``
        property in the PLX content.
    inputs:
        Inputs passed to the pipeline. Can be either a ``PipelineInputs`` dictionary
        or a ``WorkingMemory`` instance.
    output_name:
        Name of the output slot to write to.
    output_multiplicity:
        Output multiplicity specification.
    dynamic_output_concept_code:
        Override the dynamic output concept code.
    pipe_run_mode:
        Pipe run mode: ``PipeRunMode.LIVE`` or ``PipeRunMode.DRY``. If not specified,
        inferred from the environment variable ``PIPELEX_FORCE_DRY_RUN_MODE``. Defaults
        to ``PipeRunMode.LIVE`` if the environment variable is not set.
    search_domains:
        List of domains to search for pipes. The executed pipe's domain is automatically
        added if not already present.

    Returns:
    -------
    tuple[PipeJob, str, str]
        A tuple containing the pipe job ready for execution, the pipeline run ID,
        and the library ID.

    """
    if not plx_content and not pipe_code:
        msg = "Either pipe_code or plx_content must be provided to the pipeline API."
        raise ValueError(msg)

    pipeline = get_pipeline_manager().add_new_pipeline()
    pipeline_run_id = pipeline.pipeline_run_id

    if not library_id:
        library_id = pipeline_run_id

    library_manager = get_library_manager()
    set_current_library(library_id=library_id)
    library_manager.open_library(library_id=library_id)

    pipe: PipeAbstract | None = None
    blueprint: PipelexBundleBlueprint | None = None

    if plx_content:
        validate_bundle_result = await validate_bundle(plx_content=plx_content)
        library_manager.load_from_blueprints(library_id=library_id, blueprints=validate_bundle_result.blueprints)
        # For now, we only support one blueprint when given a plx_content. So blueprints is of length 1.
        blueprint = validate_bundle_result.blueprints[0]
        if pipe_code:
            pipe = get_required_pipe(pipe_code=pipe_code)
        elif blueprint.main_pipe:
            pipe = get_required_pipe(pipe_code=blueprint.main_pipe)
        else:
            msg = "No pipe code or main pipe in the PLX content provided to the pipeline API."
            raise PipeExecutionError(message=msg)
    elif pipe_code:
        if library_dirs:
            library_manager.load_libraries(library_id=library_id, library_dirs=[Path(library_dir) for library_dir in library_dirs])
        else:
            library_manager.load_libraries(library_id=library_id, library_dirs=[Path.cwd()])
        pipe = get_required_pipe(pipe_code=pipe_code)
    else:
        msg = "Either provide pipe_code or plx_content to the pipeline API. 'pipe_code' must be provided when 'plx_content' is None"
        raise PipeExecutionError(message=msg)

    search_domains = search_domains or []
    if pipe.domain not in search_domains:
        search_domains.insert(0, pipe.domain)

    working_memory: WorkingMemory | None = None

    if inputs:
        if isinstance(inputs, WorkingMemory):
            working_memory = inputs
        else:
            working_memory = WorkingMemoryFactory.make_from_pipeline_inputs(
                pipeline_inputs=inputs,
                search_domains=search_domains,
            )

    if pipe_run_mode is None:
        if run_mode_from_env := get_optional_env(key=FORCE_DRY_RUN_MODE_ENV_KEY):
            pipe_run_mode = PipeRunMode(run_mode_from_env)
        else:
            pipe_run_mode = PipeRunMode.LIVE

    get_report_delegate().open_registry(pipeline_run_id=pipeline_run_id)

    job_metadata = JobMetadata(
        pipeline_run_id=pipeline.pipeline_run_id,
    )

    pipe_run_params = PipeRunParamsFactory.make_run_params(
        output_multiplicity=output_multiplicity,
        dynamic_output_concept_code=dynamic_output_concept_code,
        pipe_run_mode=pipe_run_mode,
    )

    pipe_job = PipeJobFactory.make_pipe_job(
        pipe=pipe,
        pipe_run_params=pipe_run_params,
        job_metadata=job_metadata,
        working_memory=working_memory,
        output_name=output_name,
    )

    properties = {
        EventProperty.PIPELINE_RUN_ID: job_metadata.pipeline_run_id,
        EventProperty.PIPE_TYPE: pipe.pipe_type,
    }
    get_telemetry_manager().track_event(event_name=EventName.PIPELINE_EXECUTE, properties=properties)

    return pipe_job, pipeline_run_id, library_id



import logging

from typing_extensions import override

from pipelex.tools.log.log_levels import LOGGING_LEVEL_DEV, LOGGING_LEVEL_VERBOSE
from pipelex.tools.misc.terminal_utils import BOLD_FONT, RESET_FONT, TerminalColor


# TODO: move these to the config
def emoji_for_channel(channel_name: str) -> str | None:
    channel_emojis: dict[str, str] = {
        "root": "",
        "werkzeug": "📡",
        "urllib3.connectionpool": "⚡️",
    }

    emoji = channel_emojis.get(channel_name)
    if emoji == "":
        # blank emoji is OK
        return emoji
    elif emoji:
        return emoji
    elif channel_name.startswith("google"):
        return "🌀"
    elif channel_name.startswith("openai"):
        return "⚪️"
    elif channel_name.startswith("kajson"):
        # space added to make it look better
        return "*️⃣ "
    elif channel_name.startswith("#poor-log"):
        # space added to make it look better
        return "🧿 "
    elif channel_name.startswith("pipelex"):
        return "🧠"
    else:
        return None


class EmojiLogFormatter(logging.Formatter):
    @override
    def format(self, record: logging.LogRecord):
        log_fmt: str
        emoji = emoji_for_channel(record.name)
        if emoji == "":
            log_fmt = "%(message)s"
        elif emoji:
            log_fmt = f"{emoji}: %(message)s"
        else:
            log_fmt = "[%(name)s]: %(message)s"
        formatter = logging.Formatter(log_fmt)

        return formatter.format(record)


log_level_color: dict[int, TerminalColor] = {
    LOGGING_LEVEL_VERBOSE: TerminalColor.WHITE,
    logging.DEBUG: TerminalColor.GREEN,
    LOGGING_LEVEL_DEV: TerminalColor.CYAN,
    logging.INFO: TerminalColor.BLUE,
    logging.WARNING: TerminalColor.YELLOW,
    logging.ERROR: TerminalColor.RED,
    logging.CRITICAL: TerminalColor.MAGENTA,
}

# added spaces and truncated to 5 characters to make it look better
log_level_tag: dict[int, str] = {
    LOGGING_LEVEL_VERBOSE: "VERBO",
    logging.DEBUG: "DEBUG",
    LOGGING_LEVEL_DEV: "DEV  ",
    logging.INFO: "INFO ",
    logging.WARNING: f"{BOLD_FONT}WARNING{RESET_FONT}",
    logging.ERROR: f"{BOLD_FONT}/ERROR\\{RESET_FONT}",
    logging.CRITICAL: f"{BOLD_FONT}CRITICAL{RESET_FONT}",
}


class LevelAndEmojiLogFormatter(logging.Formatter):
    @override
    def format(self, record: logging.LogRecord):
        emoji = emoji_for_channel(record.name)
        color = log_level_color.get(record.levelno, RESET_FONT)
        tag = log_level_tag.get(record.levelno, "?????")
        if emoji:
            log_fmt = f"{emoji} {color}{tag}:{RESET_FONT} %(message)s"
        else:
            log_fmt = f"{color}{tag}:%(name)s{RESET_FONT} %(message)s"
        formatter = logging.Formatter(log_fmt)

        if record.levelno in [logging.WARNING, logging.ERROR, logging.CRITICAL]:
            record.msg = f"{color}{BOLD_FONT}{record.msg}{RESET_FONT}"

        return formatter.format(record)



from typing import Literal

import shortuuid
from typing_extensions import override

from pipelex import log
from pipelex.cogt.templating.template_category import TemplateCategory
from pipelex.core.concepts.concept_factory import ConceptFactory
from pipelex.core.concepts.native.concept_native import NativeConceptCode
from pipelex.core.memory.exceptions import WorkingMemoryStuffNotFoundError
from pipelex.core.memory.working_memory import WorkingMemory
from pipelex.core.pipes.exceptions import PipeValidationError, PipeValidationErrorType
from pipelex.core.pipes.inputs.input_requirements import InputRequirements
from pipelex.core.pipes.inputs.input_requirements_factory import InputRequirementsFactory
from pipelex.core.pipes.pipe_output import PipeOutput
from pipelex.hub import get_content_generator, get_optional_pipe, get_pipe_router, get_pipeline_tracker, get_required_pipe
from pipelex.pipe_controllers.condition.pipe_condition_details import PipeConditionDetails
from pipelex.pipe_controllers.condition.special_outcome import SpecialOutcome
from pipelex.pipe_controllers.pipe_controller import PipeController
from pipelex.pipe_run.exceptions import PipeRunError
from pipelex.pipe_run.pipe_job_factory import PipeJobFactory
from pipelex.pipe_run.pipe_run_params import PipeRunParams
from pipelex.pipeline.job_metadata import JobMetadata
from pipelex.tools.jinja2.jinja2_errors import Jinja2DetectVariablesError
from pipelex.tools.jinja2.jinja2_required_variables import detect_jinja2_required_variables

ConditionOutcomeMap = dict[str, str | SpecialOutcome]


class PipeCondition(PipeController):
    type: Literal["PipeCondition"] = "PipeCondition"
    expression: str
    outcome_map: ConditionOutcomeMap
    default_outcome: str | SpecialOutcome
    add_alias_from_expression_to: str | None = None

    @property
    def mapped_pipe_codes(self) -> set[str]:
        codes = set(self.outcome_map.values())
        if self.default_outcome:
            codes.add(self.default_outcome)
        return codes - set(SpecialOutcome.value_list())

    def _make_pipe_condition_details(self, evaluated_expression: str, chosen_pipe_code: str) -> PipeConditionDetails:
        return PipeConditionDetails(
            code=shortuuid.uuid()[:5],
            test_expression=self.expression,
            outcomes=self.outcome_map,
            default_pipe_code=self.default_outcome,
            evaluated_expression=evaluated_expression,
            chosen_pipe_code=chosen_pipe_code,
        )

    @override
    def pipe_dependencies(self) -> set[str]:
        return self.mapped_pipe_codes

    @override
    def required_variables(self) -> set[str]:
        required_variables: set[str] = set()
        # Variables from the expression/expression_template
        expression_required_variables = detect_jinja2_required_variables(
            template_category=TemplateCategory.EXPRESSION,
            template_source=self.expression,
        )
        required_variables.update(expression_required_variables)

        # Variables from the outcomes map and default_outcome
        for pipe_code in self.pipe_dependencies():
            required_variables.update(get_required_pipe(pipe_code=pipe_code).required_variables())
        return required_variables

    @override
    def needed_inputs(self, visited_pipes: set[str] | None = None) -> InputRequirements:
        if visited_pipes is None:
            visited_pipes = set()

        # If we've already visited this pipe, stop recursion
        if self.code in visited_pipes:
            return InputRequirementsFactory.make_empty()

        # Add this pipe to visited set for recursive calls
        visited_pipes_with_current = visited_pipes | {self.code}

        needed_inputs = InputRequirementsFactory.make_empty()

        # 1. Add the variables from the expression/expression_template
        required_variables = detect_jinja2_required_variables(
            template_category=TemplateCategory.EXPRESSION,
            template_source=self.expression,
        )

        for var_name in required_variables:
            if not var_name.startswith("_"):  # exclude internal variables starting with `_`
                # We don't know the concept code from just the variable name,
                # so we'll use a generic placeholder that will be validated later
                needed_inputs.add_requirement(
                    variable_name=var_name,
                    concept=ConceptFactory.make_native_concept(
                        native_concept_code=NativeConceptCode.ANYTHING,
                    ),
                )

        # 2. Add the inputs needed by all possible target pipes
        for pipe_code in self.mapped_pipe_codes:
            pipe = get_required_pipe(pipe_code=pipe_code)
            # Use the centralized recursion detection
            pipe_needed_inputs = pipe.needed_inputs(visited_pipes_with_current)

            for input_name, requirement in pipe_needed_inputs.items:
                needed_inputs.add_requirement(variable_name=input_name, concept=requirement.concept)
        return needed_inputs

    @override
    def validate_inputs_static(self):
        pass

    @override
    def validate_inputs_with_library(self):
        pass

    @override
    def validate_output_static(self):
        pass

    @override
    def validate_output_with_library(self):
        """Validate the output for the pipe condition.
        The output of the pipe condition should match the output of all the conditional pipes, and the default pipe.
        """
        for pipe_code in self.mapped_pipe_codes:
            pipe = get_required_pipe(pipe_code=pipe_code)
            if self.output.concept_string not in (
                pipe.output.concept_string,
                NativeConceptCode.DYNAMIC.concept_string,
                NativeConceptCode.ANYTHING.concept_string,
            ):
                msg = (
                    f"The output concept code '{self.output.concept_string}' of the pipe '{self.code}' is not "
                    f"matching the output concept code '{pipe.output.concept_string}' of the pipe '{pipe_code}'"
                )
                raise PipeValidationError(
                    message=msg,
                    error_type=PipeValidationErrorType.INADEQUATE_OUTPUT_CONCEPT,
                    domain=self.domain,
                    pipe_code=self.code,
                    provided_concept_code=pipe.output.concept_string,
                    required_concept_codes=[self.output.concept_string],
                )

    # TODO: Restore this validation. The problem lies with needed_inputs that construct Anything concepts.
    # @override
    # async def validate_before_run(
    #     self,
    #     job_metadata: JobMetadata,
    #     working_memory: WorkingMemory,
    #     pipe_run_params: PipeRunParams,
    #     output_name: str | None = None,
    # ):
    #     evaluated_expression = await get_content_generator().make_templated_text(
    #         context=working_memory.generate_context(),
    #         template=self.expression,
    #         template_category=TemplateCategory.EXPRESSION,
    #     )
    #     if not evaluated_expression or evaluated_expression == "None":
    #         error_msg = f"PipeCondition '{self.code}': Conditional expression returned no result"
    #         raise PipeRunError(
    #             message=error_msg,
    #             run_mode=pipe_run_params.run_mode,
    #         )

    async def _evaluate_expression(
        self,
        working_memory: WorkingMemory,
    ) -> str:
        """Evaluate the conditional expression and select the appropriate pipe.

        Args:
            working_memory: The working memory context for evaluation

        Returns:
            The evaluated expression
        """
        evaluated_expression = await get_content_generator().make_templated_text(
            context=working_memory.generate_context(),
            template=self.expression,
            template_category=TemplateCategory.EXPRESSION,
        )

        log.verbose(f"add_alias: {evaluated_expression} -> {self.add_alias_from_expression_to}")
        if self.add_alias_from_expression_to:
            working_memory.add_alias(
                alias=evaluated_expression,
                target=self.add_alias_from_expression_to,
            )

        return evaluated_expression

    @override
    async def _run_controller_pipe(
        self,
        job_metadata: JobMetadata,
        working_memory: WorkingMemory,
        pipe_run_params: PipeRunParams,
        output_name: str | None = None,
    ) -> PipeOutput:
        log.verbose(f"{self.class_name} generating a '{self.output.code}'")

        # TODO: restore pipe_layer feature
        # pipe_run_params.push_pipe_code(pipe_code=pipe_code)

        evaluated_expression = await self._evaluate_expression(working_memory=working_memory)

        # Select the outcome based on the evaluated expression
        outcome = self.outcome_map.get(evaluated_expression, self.default_outcome)

        # Handle continue case
        if SpecialOutcome.is_continue(outcome):
            log.dev(f"PipeCondition '{self.code}' continued with outcome: {outcome}. Evaluated expression: {evaluated_expression}")
            return PipeOutput(working_memory=working_memory)

        if SpecialOutcome.is_fail(outcome):
            msg = f"PipeCondition '{self.code}' failed with outcome: {outcome}. Evaluated expression: {evaluated_expression}"
            raise PipeRunError(message=msg, run_mode=pipe_run_params.run_mode)

        chosen_pipe = get_required_pipe(pipe_code=outcome)

        # Create condition details for tracking
        condition_details = self._make_pipe_condition_details(
            evaluated_expression=self.expression,
            chosen_pipe_code=chosen_pipe.code,
        )

        # Get required variables and validate they exist in working memory
        required_variables = chosen_pipe.required_variables()
        required_stuff_names = {required_variable for required_variable in required_variables if not required_variable.startswith("_")}
        try:
            required_stuffs = working_memory.get_stuffs(names=required_stuff_names)
        except WorkingMemoryStuffNotFoundError as exc:
            pipe_condition_path = [*pipe_run_params.pipe_layers, self.code]
            pipe_condition_path_str = ".".join(pipe_condition_path)
            error_details = f"PipeCondition '{pipe_condition_path_str}', required_variables: {required_variables}, missing: '{exc.variable_name}'"
            msg = f"Some required stuff(s) not found: {error_details}"
            raise PipeRunError(message=msg, run_mode=pipe_run_params.run_mode) from exc

        # Track condition steps
        for required_stuff in required_stuffs:
            get_pipeline_tracker().add_condition_step(
                from_stuff=required_stuff,
                to_condition=condition_details,
                condition_expression=self.expression,
                pipe_layer=pipe_run_params.pipe_layers,
                comment="PipeCondition required for condition",
            )

        # Execute the chosen pipe
        log.verbose(f"Chosen pipe: {chosen_pipe.code}")
        pipe_output = await get_pipe_router().run(
            pipe_job=PipeJobFactory.make_pipe_job(
                pipe=chosen_pipe,
                job_metadata=job_metadata,
                working_memory=working_memory,
                pipe_run_params=pipe_run_params,
                output_name=output_name,
            ),
        )

        # Track choice step
        get_pipeline_tracker().add_choice_step(
            from_condition=condition_details,
            to_stuff=pipe_output.main_stuff,
            pipe_layer=pipe_run_params.pipe_layers,
            comment="PipeCondition chosen pipe",
        )
        return pipe_output

    @override
    async def _dry_run_controller_pipe(
        self,
        job_metadata: JobMetadata,
        working_memory: WorkingMemory,
        pipe_run_params: PipeRunParams,
        output_name: str | None = None,
    ) -> PipeOutput:
        """Dry run implementation for PipeCondition.
        Validates that all required inputs are present, expression is valid, and target pipes exist.
        """
        log.verbose(f"PipeCondition: dry run controller pipe: {self.code}")
        # 1. Validate that all required inputs are present in the working memory
        needed_inputs = self.needed_inputs()
        missing_input_names: list[str] = []

        for named_input_requirement in needed_inputs.named_input_requirements:
            if not working_memory.get_optional_stuff(named_input_requirement.variable_name):
                missing_input_names.append(named_input_requirement.variable_name)

        if missing_input_names:
            log.error(f"Dry run failed: missing required inputs: {missing_input_names}")
            raise PipeRunError(
                message=f"Dry run failed for pipe '{self.code}' (PipeCondition): missing required inputs: {', '.join(missing_input_names)}",
                run_mode=pipe_run_params.run_mode,
            )

        # 2. Validate that the expression template is valid
        try:
            required_variables = detect_jinja2_required_variables(
                template_category=TemplateCategory.EXPRESSION,
                template_source=self.expression,
            )
            log.verbose(f"Expression template is valid, requires variables: {required_variables}")
        except Jinja2DetectVariablesError as exc:
            log.error(f"Dry run failed: could not detect required variables from expression template: {exc}")
            msg = (
                f"Dry run failed for pipe '{self.code}' (PipeCondition): could not detect required variables "
                f"from expression template: {exc}\nTemplate:\n'{self.expression}'"
            )
            raise PipeRunError(message=msg, run_mode=pipe_run_params.run_mode) from exc

        # 3. Validate that all values in the outcomes map (appart from special outcomes) do exist as pipe codes
        all_pipe_codes = set(self.outcome_map.values())
        if self.default_outcome:
            all_pipe_codes.add(self.default_outcome)
        all_pipe_codes -= set(SpecialOutcome.value_list())

        missing_pipes = [pipe_code for pipe_code in all_pipe_codes if not get_optional_pipe(pipe_code=pipe_code)]

        if missing_pipes:
            msg = (
                f"Dry run failed for PipeCondition '{self.code}': missing pipes: {', '.join(missing_pipes)}. "
                f"Pipe map: {self.outcome_map}, default: {self.default_outcome}"
            )
            raise PipeRunError(message=msg, run_mode=pipe_run_params.run_mode)

        # Here, it should launch the dry run of all the pipes in the outcomes map
        for pipe_code in self.mapped_pipe_codes:
            pipe = get_required_pipe(pipe_code=pipe_code)
            await pipe.run_pipe(
                job_metadata=job_metadata,
                working_memory=working_memory,
                pipe_run_params=pipe_run_params,
            )
        return PipeOutput(working_memory=working_memory)



from typing import Any

from pydantic import ValidationError

from pipelex.cogt.exceptions import ModelDeckNotFoundError, ModelDeckValidationError
from pipelex.cogt.models.model_deck import ModelDeckBlueprint
from pipelex.tools.misc.json_utils import deep_update
from pipelex.tools.misc.toml_utils import load_toml_from_path
from pipelex.tools.typing.pydantic_utils import format_pydantic_validation_error


def load_model_deck_blueprint(model_deck_paths: list[str]) -> ModelDeckBlueprint:
    full_deck_dict: dict[str, Any] = {}
    if not model_deck_paths:
        msg = "No Model deck paths found. Please run `pipelex init config` to create the set up the base deck."
        raise ModelDeckNotFoundError(msg)

    for deck_path in model_deck_paths:
        try:
            deck_dict = load_toml_from_path(path=deck_path)
        except FileNotFoundError as not_found_exc:
            msg = f"Could not find Model Deck file at '{deck_path}': {not_found_exc}"
            raise ModelDeckNotFoundError(msg) from not_found_exc
        deep_update(full_deck_dict, deck_dict)

    try:
        return ModelDeckBlueprint.model_validate(full_deck_dict)
    except ValidationError as exc:
        valiation_error_msg = format_pydantic_validation_error(exc)
        msg = f"Invalid Model Deck configuration in {model_deck_paths}: {valiation_error_msg}"
        raise ModelDeckValidationError(msg) from exc



from typing import TYPE_CHECKING, Literal

from pydantic import Field, field_validator
from rich.console import Group
from rich.text import Text
from typing_extensions import override

from pipelex.builder.pipe.pipe_spec import PipeSpec
from pipelex.pipe_operators.img_gen.pipe_img_gen_blueprint import PipeImgGenBlueprint
from pipelex.tools.misc.pretty import PrettyPrintable
from pipelex.types import StrEnum

if TYPE_CHECKING:
    from pipelex.cogt.img_gen.img_gen_setting import ImgGenModelChoice


class ImgGenSkill(StrEnum):
    GEN_IMAGE_BASIC = "gen_image_basic"
    GEN_IMAGE_FAST = "gen_image_fast"
    GEN_IMAGE_HIGH_QUALITY = "gen_image_high_quality"


class PipeImgGenSpec(PipeSpec):
    """Specs for image generation pipe operations in the Pipelex framework.

    PipeImgGen enables AI-powered image generation using various models like DALL-E or
    diffusion models. Supports static and dynamic prompts with configurable generation
    parameters.

    Output Multiplicity:
        Specify using bracket notation in output field:
        - output = "Image" - single image (default)
        - output = "Image[3]" - exactly 3 images
    """

    type: Literal["PipeImgGen"] = "PipeImgGen"
    pipe_category: Literal["PipeOperator"] = "PipeOperator"
    img_gen_skill: ImgGenSkill | str = Field(description="Select the most adequate image generation skill according to the task to be performed.")

    @field_validator("img_gen_skill", mode="before")
    @classmethod
    def validate_img_gen_skill(cls, img_gen_skill_value: str | None) -> ImgGenSkill | None:
        if img_gen_skill_value is None:
            return None
        else:
            return ImgGenSkill(img_gen_skill_value)

    @override
    def rendered_pretty(self, title: str | None = None, depth: int = 0) -> PrettyPrintable:
        # Get base pipe information from parent
        base_group = super().rendered_pretty(title=title, depth=depth)

        # Create a group combining base info with img_gen-specific details
        img_gen_group = Group()
        img_gen_group.renderables.append(base_group)

        # Add image generation specific information
        img_gen_group.renderables.append(Text())  # Blank line
        img_gen_group.renderables.append(Text.from_markup(f"Image Generation Skill: [bold yellow]{self.img_gen_skill}[/bold yellow]"))

        return img_gen_group

    @override
    def to_blueprint(self) -> PipeImgGenBlueprint:
        """Convert this PipeImgGenBlueprint to the core PipeImgGenBlueprint."""
        base_blueprint = super().to_blueprint()

        img_gen_choice: ImgGenModelChoice = self.img_gen_skill

        return PipeImgGenBlueprint(
            description=base_blueprint.description,
            inputs=base_blueprint.inputs,
            output=base_blueprint.output,
            img_gen_prompt=None,
            img_gen_prompt_var_name=base_blueprint.input_names[0] if base_blueprint.input_names else None,
            model=img_gen_choice,
            aspect_ratio=None,
            background=None,
            output_format=None,
            is_raw=None,
            seed=None,
        )



from typing import Literal, Union

from pydantic import Field, field_validator

from pipelex.cogt.llm.llm_job_components import LLMJobParams
from pipelex.cogt.model_backends.prompting_target import PromptingTarget
from pipelex.system.configuration.config_model import ConfigModel
from pipelex.types import Self


class LLMSetting(ConfigModel):
    model: str
    temperature: float = Field(..., ge=0, le=1)
    max_tokens: int | None = None
    prompting_target: PromptingTarget | None = Field(default=None, strict=False)

    @field_validator("max_tokens", mode="before")
    @classmethod
    def validate_max_tokens(cls, value: int | Literal["auto"] | None) -> int | None:
        if value is None or (isinstance(value, str) and value == "auto"):
            return None
        if isinstance(value, int):  # pyright: ignore[reportUnnecessaryIsInstance]
            return value

    def make_llm_job_params(self) -> LLMJobParams:
        return LLMJobParams(
            temperature=self.temperature,
            max_tokens=self.max_tokens,
            seed=None,
        )

    def desc(self) -> str:
        return (
            f"LLMSetting(llm_handle={self.model}, temperature={self.temperature}, "
            f"max_tokens={self.max_tokens}, prompting_target={self.prompting_target})"
        )


LLMModelChoice = Union[LLMSetting, str]


class LLMSettingChoicesDefaults(ConfigModel):
    for_text: LLMModelChoice
    for_object: LLMModelChoice


class LLMSettingChoices(ConfigModel):
    for_text: LLMModelChoice | None
    for_object: LLMModelChoice | None

    def list_choice_strings(self) -> set[str]:
        return {c for c in (self.for_text, self.for_object) if isinstance(c, str)}

    @classmethod
    def make_completed_with_defaults(
        cls,
        for_text: LLMModelChoice | None = None,
        for_object: LLMModelChoice | None = None,
    ) -> Self:
        return cls(
            for_text=for_text,
            for_object=for_object,
        )



from pydantic import RootModel
from typing_extensions import override

from pipelex.core.domains.domain import Domain
from pipelex.libraries.domain.domain_library_abstract import DomainLibraryAbstract
from pipelex.libraries.domain.exceptions import DomainLibraryError
from pipelex.types import Self

DomainLibraryRoot = dict[str, Domain]


class DomainLibrary(RootModel[DomainLibraryRoot], DomainLibraryAbstract):
    def setup(self):
        pass

    @override
    def teardown(self):
        self.root = {}

    def reset(self):
        self.teardown()
        self.setup()

    @classmethod
    def make_empty(cls) -> Self:
        return cls(root={})

    @override
    def get_domain(self, domain: str) -> Domain | None:
        return self.root.get(domain)

    def add_domain(self, domain: Domain):
        domain_code = domain.code
        if domain_code in self.root:
            msg = f"Trying to add domain '{domain_code}' to domain library but it already exists"
            raise DomainLibraryError(msg)
        self.root[domain_code] = domain

    def add_domains(self, domains: list[Domain]):
        for domain in domains:
            self.add_domain(domain=domain)

    def remove_domain_by_code(self, domain_code: str) -> None:
        if domain_code not in self.root:
            msg = f"Trying to remove domain '{domain_code}' from domain library but it does not exist"
            raise DomainLibraryError(msg)
        del self.root[domain_code]

    @override
    def get_required_domain(self, domain: str) -> Domain:
        """Get a domain by code from this library, raising an error if not found."""
        the_domain = self.get_domain(domain=domain)
        if not the_domain:
            msg = f"Domain '{domain}' not found. Check for typos and make sure it is declared in a pipeline library."
            raise DomainLibraryError(msg)
        return the_domain



import json
from collections.abc import Mapping
from typing import Any, Union, cast

from kajson import kajson
from pydantic import BaseModel

from pipelex.system.exceptions import ToolError
from pipelex.tools.misc.file_utils import save_text_to_path
from pipelex.tools.typing.pydantic_utils import CustomBaseModel

JsonContent = Union[dict[str, Any], list[Any]]


class ArgumentTypeError(ToolError):
    pass


class JsonTypeError(ToolError):
    pass


def json_str(some_object: Any, title: str | None = None, is_spaced: bool = False) -> str:
    """Creates a formatted JSON string representation of any Python object with optional title and spacing.

    This function is a higher-level wrapper around purify_json that provides additional formatting
    options. It always uses 4-space indentation and disables warning wrapping for non-serializable
    data.

    Args:
        some_object (Any): The object to convert to a JSON string. Can be any type supported
            by purify_json.
        title (str | None, optional): A title to prepend to the JSON string. If provided,
            the output will be in the format "title: {json_string}". Defaults to None.
        is_spaced (bool, optional): If True, adds newlines before and after the JSON string
            for better readability. Defaults to False.

    Returns:
        str: The formatted JSON string representation of the object.

    Example:
        >>> data = {"name": "test", "values": [1, 2, 3]}
        >>> print(json_str(data, title="Data", is_spaced=True))

        Data: {
            "name": "test",
            "values": [1, 2, 3]
        }

    """
    _, json_string = purify_json(some_object, indent=4, is_warning_enabled=False)
    if title:
        json_string = f"{title}: {json_string}"
    if is_spaced:
        json_string = f"\n{json_string}\n"
    return json_string


def save_as_json_to_path(
    object_to_save: Any,
    path: str,
    indent: int | None = 4,
    is_warning_enabled: bool = True,
):
    """Saves a Python object as a JSON file at the specified path.

    This function converts a Python object to a JSON string and saves it to a file. The object
    is first purified to ensure JSON compatibility before saving.

    Args:
        object_to_save (Any): The Python object to be saved as JSON. Can be any JSON-serializable object.
        path (str): The file path where the JSON file will be saved.
        indent (int | None, optional): Number of spaces for JSON formatting indentation. Defaults to 4.
        is_warning_enabled (bool, optional): Whether to show warnings during JSON purification. Defaults to True.

    Returns:
        None

    """
    _, json_string = purify_json(object_to_save, indent=indent, is_warning_enabled=is_warning_enabled)
    save_text_to_path(json_string, path)


def load_json_from_path(path: str) -> JsonContent:
    """Loads and parses a JSON file from the specified path.

    This function reads a JSON file and returns its contents as a Python object.
    The file is read using UTF-8 encoding.

    Args:
        path (str): The file path to the JSON file to be loaded.

    Returns:
        JsonContent: The parsed JSON content as a Python object (can be a dict, list, string, number, bool, or None).

    Raises:
        FileNotFoundError: If the file does not exist.
        json.JSONDecodeError: If the file contains invalid JSON.

    """
    with open(path, encoding="utf-8") as file:
        json_content: JsonContent = json.load(file)
        return json_content


def load_json_dict_from_path(path: str) -> dict[Any, Any]:
    """Loads a JSON file and ensures it contains a dictionary.

    This function reads a JSON file and verifies that its content is a dictionary.
    It uses load_json_from_path internally and adds type checking.

    Args:
        path (str): The file path to the JSON file to be loaded.

    Returns:
        Dict[Any, Any]: The parsed JSON content as a Python dictionary.

    Raises:
        JsonTypeError: If the JSON content is not a dictionary.
        FileNotFoundError: If the file does not exist.
        json.JSONDecodeError: If the file contains invalid JSON.

    """
    json_content: JsonContent = load_json_from_path(path)
    if not isinstance(json_content, dict):
        msg = f"{path} is not a dict"
        raise JsonTypeError(msg)
    return json_content


def load_json_list_from_path(path: str) -> list[Any]:
    """Loads a JSON file and ensures it contains a list.

    This function reads a JSON file and verifies that its content is a list.
    It uses load_json_from_path internally and adds type checking.

    Args:
        path (str): The file path to the JSON file to be loaded.

    Returns:
        List[Any]: The parsed JSON content as a Python list.

    Raises:
        JsonTypeError: If the JSON content is not a list.
        FileNotFoundError: If the file does not exist.
        json.JSONDecodeError: If the file contains invalid JSON.

    """
    json_content: JsonContent = load_json_from_path(path)
    if isinstance(json_content, list):
        return json_content
    msg = f"{path} is not a list"
    raise JsonTypeError(msg)


def deep_update(target_dict: dict[str, Any], updates: dict[str, Any]):
    """Recursively updates a dictionary with values from another dictionary.

    This function performs a deep merge of two dictionaries, handling nested
    dictionaries. For dictionaries, it recursively updates values. For all other
    types (including lists), values from updates override the target values.

    Args:
        target_dict (Dict[str, Any]): The dictionary to update. This dictionary
            will be modified in place.
        updates (Dict[str, Any]): The dictionary containing updates to apply.

    Example:
        >>> base = {"a": 1, "b": {"x": 2, "y": 3}, "c": [1, 2]}
        >>> updates = {"b": {"y": 4, "z": 5}, "c": [3, 4]}
        >>> deep_update(base, updates)
        >>> print(base)
        {'a': 1, 'b': {'x': 2, 'y': 4, 'z': 5}, 'c': [3, 4]}

    """
    for key, value in updates.items():
        if isinstance(value, dict) and key in target_dict and isinstance(target_dict[key], dict):
            deep_update(target_dict[key], value)  # pyright: ignore[reportUnknownArgumentType]
        else:
            target_dict[key] = value


def remove_none_values(json_content: JsonContent | Any) -> JsonContent | Any:
    """Recursively removes all None values from a JSON-compatible data structure.

    This function traverses dictionaries and lists, removing any keys with None values
    from dictionaries and processing nested structures. It preserves the structure of
    the input while cleaning out None values at any depth.

    Args:
        json_content (JsonContent): The JSON-compatible data structure (dict or list)
            to process. Can contain nested dictionaries and lists.

    Returns:
        JsonContent: A new data structure of the same type as the input, but with all
            None values removed from dictionaries at any level of nesting.

    Example:
        >>> data = {
        ...     "name": "test",
        ...     "value": None,
        ...     "nested": {"key": "exists", "empty": None},
        ...     "list": [1, None, {"key": None}]
        ... }
        >>> result = remove_none_values(data)
        >>> print(result)
        {
            "name": "test",
            "nested": {"key": "exists"},
            "list": [1, {}]
        }

    """
    if isinstance(json_content, dict):
        json_content = cast("dict[str, Any]", json_content)  # pyright: ignore[reportUnnecessaryCast]
        cleaned_dict: dict[str, Any] = {}
        for key, value in json_content.items():
            if value is not None:
                cleaned_dict[key] = remove_none_values(json_content=value)
        return cleaned_dict
    elif isinstance(json_content, list):
        json_content = cast("list[Any]", json_content)  # pyright: ignore[reportUnnecessaryCast]
        return [remove_none_values(item) for item in json_content]
    else:
        return json_content


def remove_none_values_from_dict(data: Mapping[str, Any]) -> dict[str, Any]:
    processed = remove_none_values(json_content=data)
    if not isinstance(processed, dict):
        msg = "Removing None values from a dict, we expected a dict in return"
        raise JsonTypeError(msg)
    return cast("dict[str, Any]", processed)  # pyright: ignore[reportUnnecessaryCast]


def purify_json(
    data: Any,
    indent: int | None = None,
    is_truncate_bytes_enabled: bool = False,
    is_warning_enabled: bool = True,
) -> tuple[dict[Any, Any] | list[Any], str]:
    """Converts any Python object into a JSON-serializable format and its string representation.

    This function handles various types of input data:
    - Pydantic BaseModel instances are converted using their model_dump method
    - Lists of BaseModel instances are converted element by element
    - Other types are attempted to be JSON serialized directly
    - If standard JSON serialization fails, it attempts using kajson or falls back to str conversion

    Args:
        data (Any): The data to convert. Can be a Pydantic model, list, dict, or any other type.
        indent (int | None, optional): Number of spaces for JSON formatting indentation. Defaults to None.
        is_truncate_bytes_enabled (bool, optional): If True, truncates bytes values to a string representation. Defaults to False.
        is_warning_enabled (bool, optional): If True, wraps non-serializable data in a warning object.
            Defaults to True.

    Returns:
        Tuple[Union[Dict[Any, Any], List[Any]], str]: A tuple containing:
            - The purified data structure (either a dict or list)
            - The JSON string representation of the data

    Example:
        >>> model = SomeModel(name="test")
        >>> data, json_str = purify_json(model)
        >>> print(json_str)
        '{"name": "test"}'

    """
    dict_string: str
    if isinstance(data, CustomBaseModel) and is_truncate_bytes_enabled:
        return purify_json(
            data.model_dump_truncated(serialize_as_any=True),
            indent=indent,
            is_truncate_bytes_enabled=is_truncate_bytes_enabled,
            is_warning_enabled=is_warning_enabled,
        )
    if isinstance(data, BaseModel):
        return purify_json(
            data.model_dump(serialize_as_any=True),
            indent=indent,
            is_truncate_bytes_enabled=is_truncate_bytes_enabled,
            is_warning_enabled=is_warning_enabled,
        )

    if isinstance(data, list):
        the_list = data  # pyright: ignore[reportUnknownVariableType]
        if not the_list:
            return [], "[]"
        if isinstance(the_list[0], CustomBaseModel) and is_truncate_bytes_enabled:
            the_list_of_custom_base_models = cast("list[CustomBaseModel]", the_list)
            pure_list = [item.model_dump_truncated(serialize_as_any=True) for item in the_list_of_custom_base_models]
            dict_string = json.dumps(pure_list, indent=indent, default=str)
            return pure_list, dict_string
        if isinstance(the_list[0], BaseModel):
            the_list_of_base_models = cast("list[BaseModel]", the_list)
            pure_list = [item.model_dump(serialize_as_any=True) for item in the_list_of_base_models]
            dict_string = json.dumps(pure_list, indent=indent, default=str)
            return pure_list, dict_string

    try:
        dict_string = json.dumps(data, indent=indent)
        pure_dict = cast("dict[Any, Any] | list[Any]", data)
    except TypeError:
        try:
            dict_string = kajson.dumps(data, indent=indent)  # pyright: ignore[reportUnknownMemberType]
        except Exception:
            if is_warning_enabled:
                data = cast("dict[Any, Any] | list[Any]", data)
                data = {"!": data}
            dict_string = json.dumps(data, indent=indent, default=str)
        pure_dict = json.loads(dict_string)
    return pure_dict, dict_string


def purify_json_list(
    data: list[Any],
    indent: int | None = None,
    is_truncate_bytes_enabled: bool = False,
) -> tuple[list[Any], str]:
    """Converts a list of Python objects into a JSON-serializable list and its string representation.

    This function specifically handles lists and provides specialized processing for lists of
    Pydantic BaseModel instances. It attempts multiple serialization methods to ensure successful
    conversion.

    Args:
        data (List[Any]): The list to convert. Can contain Pydantic models or other types.
        indent (int | None, optional): Number of spaces for JSON formatting indentation.
            Defaults to None.
        is_truncate_bytes_enabled (bool, optional): If True, truncates bytes values to a string representation. Defaults to False.

    Returns:
        Tuple[List[Any], str]: A tuple containing:
            - The purified list with JSON-serializable elements
            - The JSON string representation of the list

    Example:
        >>> models = [SomeModel(name="test1"), SomeModel(name="test2")]
        >>> data, json_str = purify_json_list(models)
        >>> print(json_str)
        '[{"name": "test1"}, {"name": "test2"}]'

    """
    list_string: str
    pure_list: list[Any]

    if not data:
        return [], "[]"
    if isinstance(data[0], CustomBaseModel) and is_truncate_bytes_enabled:
        the_list_of_custom_base_models = cast("list[CustomBaseModel]", data)
        pure_list = [item.model_dump_truncated(serialize_as_any=True) for item in the_list_of_custom_base_models]
        list_string = json.dumps(pure_list, indent=indent, default=str)
        return pure_list, list_string
    if isinstance(data[0], BaseModel):
        the_list_of_base_models: list[BaseModel] = data
        pure_list = [item.model_dump(serialize_as_any=True) for item in the_list_of_base_models]
        list_string = json.dumps(pure_list, indent=indent, default=str)
        return pure_list, list_string

    try:
        list_string = json.dumps(data, indent=indent)
        pure_list = data
    except TypeError:
        try:
            list_string = kajson.dumps(data, indent=indent)  # pyright: ignore[reportUnknownMemberType]
        except Exception:
            list_string = json.dumps(data, indent=indent, default=str)
        pure_list = json.loads(list_string)
    return pure_list, list_string


def purify_json_dict(data: Any, indent: int | None = None, is_warning_enabled: bool = True) -> tuple[dict[str, Any], str]:
    """Converts any Python object into a JSON-serializable dictionary and its string representation.

    This function specifically handles dictionary-like objects and Pydantic BaseModel instances,
    converting them to pure dictionaries that can be JSON serialized. It includes multiple
    fallback mechanisms for handling non-standard JSON types.

    Args:
        data (Any): The data to convert. Can be a Pydantic model or dictionary-like object.
        indent (int | None, optional): Number of spaces for JSON formatting indentation.
            Defaults to None.
        is_warning_enabled (bool, optional): If True, wraps non-serializable data in a warning
            object with a "!" key. Defaults to True.

    Returns:
        Tuple[Dict[str, Any], str]: A tuple containing:
            - The purified dictionary with JSON-serializable values
            - The JSON string representation of the dictionary

    Raises:
        ArgumentTypeError: If the input data is a list instead of a dictionary-like object.

    Example:
        >>> model = SomeModel(name="test", value=123)
        >>> data, json_str = purify_json_dict(model)
        >>> print(json_str)
        '{"name": "test", "value": 123}'

    """
    dict_string: str
    if isinstance(data, BaseModel):
        return purify_json_dict(
            data.model_dump(serialize_as_any=True),
            indent=indent,
            is_warning_enabled=is_warning_enabled,
        )

    if isinstance(data, list):
        msg = "The data is a list, not a dict"
        raise ArgumentTypeError(msg)

    try:
        dict_string = json.dumps(data, indent=indent)
        pure_dict: dict[str, Any] = data
    except TypeError:
        try:
            dict_string = kajson.dumps(data, indent=indent)  # pyright: ignore[reportUnknownMemberType]
        except Exception:
            if is_warning_enabled:
                data = {"!": data}
            dict_string = json.dumps(data, indent=indent, default=str)
        pure_dict = json.loads(dict_string)
    return pure_dict, dict_string


def pure_json_str(data: Any, indent: int | None = None, is_warning_enabled: bool = True) -> str:
    """Converts any Python object directly to its JSON string representation.

    This is a convenience wrapper around purify_json that returns only the string
    representation, discarding the intermediate data structure. It inherits all the
    type handling capabilities of purify_json.

    Args:
        data (Any): The data to convert to a JSON string. Can be any type supported
            by purify_json.
        indent (int | None, optional): Number of spaces for JSON formatting indentation.
            Defaults to None.
        is_warning_enabled (bool, optional): If True, wraps non-serializable data in a
            warning object. Defaults to True.

    Returns:
        str: The JSON string representation of the data.

    Example:
        >>> model = SomeModel(name="test")
        >>> json_str = pure_json_str(model, indent=2)
        >>> print(json_str)
        '{
          "name": "test"
        }'

    """
    _, json_string = purify_json(data, indent=indent, is_warning_enabled=is_warning_enabled)
    return json_string



from typing import Any

from pydantic import Field

from pipelex.cogt.model_backends.model_spec import InferenceModelSpec
from pipelex.system.configuration.config_model import ConfigModel


class InferenceBackend(ConfigModel):
    name: str
    display_name: str | None = None
    enabled: bool = True
    endpoint: str | None = None
    api_key: str | None = None
    extra_config: dict[str, Any] = Field(default_factory=dict)
    model_specs: dict[str, InferenceModelSpec] = Field(default_factory=dict)

    def list_model_names(self) -> list[str]:
        """List the names of all models in the backend."""
        return list(self.model_specs.keys())

    def get_model_spec(self, model_name: str) -> InferenceModelSpec | None:
        """Get a model spec by name."""
        return self.model_specs.get(model_name)

    def get_extra_config(self, key: str) -> Any | None:
        """Get an extra config by key."""
        return self.extra_config.get(key)



def find_span(text: str, begin: str, end: str) -> tuple[int, int] | None:
    """Find the span of text between begin and end markers.

    Args:
        text: Text to search in
        begin: Beginning marker string
        end: Ending marker string

    Returns:
        Tuple of (start, end) indices if both markers found, None otherwise
        The end index includes the end marker itself
    """
    start = text.find(begin)
    if start == -1:
        return None

    end_pos = text.find(end, start)
    if end_pos == -1:
        return None

    end_pos += len(end)
    return (start, end_pos)


def wrap(begin: str, end: str, content: str) -> str:
    """Wrap content with begin and end markers.

    Args:
        begin: Beginning marker
        end: Ending marker
        content: Content to wrap

    Returns:
        Wrapped content with markers and newlines
    """
    return f"{begin}\n{content.rstrip()}\n{end}"


def replace_span(text: str, span: tuple[int, int], replacement: str) -> str:
    """Replace the text at the given span with replacement.

    Args:
        text: Original text
        span: Tuple of (start, end) indices
        replacement: Replacement text

    Returns:
        Text with span replaced by replacement
    """
    start, end = span
    return text[:start] + replacement + text[end:]



from pipelex.base_exceptions import PipelexError


class PipeRunInputsError(PipelexError):
    def __init__(self, message: str, pipe_code: str, missing_inputs: dict[str, str]):
        self.pipe_code = pipe_code
        self.missing_inputs = missing_inputs
        super().__init__(message)


class PipeInputError(PipelexError):
    def __init__(self, message: str, pipe_code: str, variable_name: str, concept_code: str | None = None):
        self.pipe_code = pipe_code
        self.variable_name = variable_name
        self.concept_code = concept_code
        super().__init__(message)


class PipeInputNotFoundError(PipelexError):
    pass



from typing import Any, cast

from pydantic import ConfigDict, ValidationError
from typing_extensions import override

from pipelex import log
from pipelex.core.concepts.concept import Concept
from pipelex.core.stuffs.exceptions import StuffArtefactReservedFieldError, StuffContentTypeError, StuffContentValidationError
from pipelex.core.stuffs.html_content import HtmlContent
from pipelex.core.stuffs.image_content import ImageContent
from pipelex.core.stuffs.list_content import ListContent
from pipelex.core.stuffs.mermaid_content import MermaidContent
from pipelex.core.stuffs.number_content import NumberContent
from pipelex.core.stuffs.pdf_content import PDFContent
from pipelex.core.stuffs.stuff_artefact import BaseStuffArtefactField, StuffArtefact
from pipelex.core.stuffs.stuff_content import StuffContent, StuffContentType
from pipelex.core.stuffs.text_and_images_content import TextAndImagesContent
from pipelex.core.stuffs.text_content import TextContent
from pipelex.tools.misc.pretty import PrettyPrintable, PrettyRenderable
from pipelex.tools.misc.string_utils import pascal_case_to_snake_case
from pipelex.tools.typing.pydantic_utils import CustomBaseModel, format_pydantic_validation_error


class Stuff(PrettyRenderable, CustomBaseModel):
    model_config = ConfigDict(extra="forbid", strict=True)

    stuff_code: str
    stuff_name: str | None = None
    concept: Concept
    content: StuffContent

    def make_artefact(self) -> StuffArtefact:
        artefact_dict: dict[str, Any] = self.content.model_dump(serialize_as_any=True)

        def set_artefact_field(key: str, value: str | StuffContent | None):
            if value is None:
                return
            if key in artefact_dict:
                stuff_name = self.stuff_name or f"unnamed using concept code {self.concept.code}"
                msg = f"""Cannot create stuff artefact for stuff '{stuff_name}' of concept '{self.concept.code}' because reserved field '{key}'
in the structured output '{self.content.__class__.__name__}' already exists in the stuff content.
Forbidden fields are: 'stuff_name', 'content_class', 'concept_code', 'stuff_code', 'content'."""
                raise StuffArtefactReservedFieldError(message=msg)
            artefact_dict[key] = value

        set_artefact_field(BaseStuffArtefactField.STUFF_NAME, self.stuff_name)
        set_artefact_field(BaseStuffArtefactField.CONTENT_CLASS, self.content.__class__.__name__)
        set_artefact_field(BaseStuffArtefactField.CONCEPT_CODE, self.concept.code)
        set_artefact_field(BaseStuffArtefactField.STUFF_CODE, self.stuff_code)
        set_artefact_field(BaseStuffArtefactField.CONTENT, self.content)
        return StuffArtefact(artefact_dict)

    @classmethod
    def make_stuff_name(cls, concept: Concept) -> str:
        return pascal_case_to_snake_case(name=concept.code)

    @property
    def title(self) -> str:
        name_from_concept = Stuff.make_stuff_name(concept=self.concept)
        concept_display = Concept.sentence_from_concept(concept=self.concept)
        if self.is_list:
            return f"List of [{concept_display}]"
        elif self.stuff_name:
            if self.stuff_name == name_from_concept:
                return concept_display
            else:
                return f"{self.stuff_name} (a {concept_display})"
        else:
            return concept_display

    @property
    def short_desc(self) -> str:
        return f"""{self.stuff_code}:
{self.concept.code} — {type(self.content).__name__}:
{self.content.short_desc}"""

    @override
    def __str__(self) -> str:
        return f"{self.title}\n{self.content.rendered_json()}"

    @property
    def is_list(self) -> bool:
        return isinstance(self.content, ListContent)

    @property
    def is_image(self) -> bool:
        return isinstance(self.content, ImageContent)

    @property
    def is_pdf(self) -> bool:
        return isinstance(self.content, PDFContent)

    @property
    def is_text(self) -> bool:
        return isinstance(self.content, TextContent)

    @property
    def is_number(self) -> bool:
        return isinstance(self.content, NumberContent)

    def content_as(self, content_type: type[StuffContentType]) -> StuffContentType:
        """Get content with proper typing if it's of the expected type."""
        return self.verify_content_type(self.content, content_type)

    @classmethod
    def verify_content_type(cls, content: StuffContent, content_type: type[StuffContentType]) -> StuffContentType:
        """Verify and convert content to the expected type."""
        # First try the direct isinstance check for performance
        if isinstance(content, content_type):
            return content

        # If isinstance failed, try model validation approach
        try:
            # Check if class names match (quick filter before attempting validation)
            if type(content).__name__ == content_type.__name__:
                content_dict = content.smart_dump()
                validated_content = content_type.model_validate(content_dict)
                log.verbose(f"Model validation passed: converted {type(content).__name__} to {content_type.__name__}")
                return validated_content
        except ValidationError as exc:
            formatted_error = format_pydantic_validation_error(exc)
            raise StuffContentValidationError(
                original_type=type(content).__name__,
                target_type=content_type.__name__,
                validation_error=formatted_error,
            ) from exc

        actual_type = type(content)
        msg = f"Content is of type '{actual_type}', instead of the expected '{content_type}'"
        raise StuffContentTypeError(message=msg, expected_type=content_type.__name__, actual_type=actual_type.__name__)

    def as_list_content(self) -> ListContent:  # pyright: ignore[reportMissingTypeArgument, reportUnknownParameterType]
        """Get content as ListContent with items of any type."""
        return self.content_as(content_type=ListContent)  # pyright: ignore[reportUnknownVariableType]

    def as_list_of_fixed_content_type(self, item_type: type[StuffContentType]) -> ListContent[StuffContentType]:
        """Get content as ListContent with items of type T.

        Args:
            item_type: The expected type of items in the list.

        Returns:
            A typed ListContent[StuffContentType] with proper type information

        Raises:
            TypeError: If content is not ListContent or items don't match expected type

        """
        list_content = cast("ListContent[StuffContentType]", self.content_as(content_type=ListContent))

        # Validate all items are of the expected type
        for item in list_content.items:
            self.verify_content_type(item, item_type)

        return list_content

    @property
    def as_text(self) -> TextContent:
        """Get content as TextContent if applicable."""
        return self.content_as(content_type=TextContent)

    @property
    def as_str(self) -> str:
        """Get content as string if applicable."""
        return self.as_text.text

    @property
    def as_image(self) -> ImageContent:
        """Get content as ImageContent if applicable."""
        return self.content_as(content_type=ImageContent)

    @property
    def as_pdf(self) -> PDFContent:
        """Get content as PDFContent if applicable."""
        return self.content_as(content_type=PDFContent)

    @property
    def as_text_and_image(self) -> TextAndImagesContent:
        """Get content as TextAndImageContent if applicable."""
        return self.content_as(content_type=TextAndImagesContent)

    @property
    def as_number(self) -> NumberContent:
        """Get content as NumberContent if applicable."""
        return self.content_as(content_type=NumberContent)

    @property
    def as_html(self) -> HtmlContent:
        """Get content as HtmlContent if applicable."""
        return self.content_as(content_type=HtmlContent)

    @property
    def as_mermaid(self) -> MermaidContent:
        """Get content as MermaidContent if applicable."""
        return self.content_as(MermaidContent)

    @override
    def rendered_pretty(self, title: str | None = None, depth: int = 0) -> PrettyPrintable:
        """Render stuff for pretty printing.

        Args:
            title: Optional title for the rendering
            depth: Current nesting depth, used to prevent nesting too many sub-tables which would end up too narrow in the console
        """
        if title and self.stuff_name:
            title = f"[cyan]{title}:[/cyan] — {self.stuff_name} ([bold green]{self.concept.code}[/bold green]"
        elif self.stuff_name:
            title = f"[cyan]{self.stuff_name}[/cyan] ([bold green]{self.concept.code}[/bold green])"
        elif title:
            title = f"[cyan]{title}:[/cyan] some stuff ([bold green]{self.concept.code}[/bold green])"
        else:
            title = f"Some stuff ([bold green]{self.concept.code}[/bold green])"
        return self.content.rendered_pretty(title=title, depth=depth)

    def pretty_print_stuff(self, title: str | None = None) -> None:
        title = title or f"[cyan]{self.stuff_name}[/cyan] ([bold green]{self.concept.code}[/bold green])"
        self.content.pretty_print_content(title=title)


class DictStuff(CustomBaseModel):
    """Stuff with content as dict[str, Any] instead of StuffContent.

    This is used for serialization where the content needs to be a plain dict.
    Has the exact same structure as Stuff but with dict content.
    """

    model_config = ConfigDict(extra="forbid", strict=True)
    concept: str
    content: Any



from anthropic import AsyncAnthropic
from anthropic.types import ModelInfo

from pipelex.cogt.model_backends.backend import InferenceBackend
from pipelex.plugins.anthropic.anthropic_exceptions import AnthropicModelListingError, AnthropicSDKUnsupportedError
from pipelex.plugins.anthropic.anthropic_factory import AnthropicFactory
from pipelex.plugins.plugin import Plugin


async def anthropic_list_available_models(plugin: Plugin, backend: InferenceBackend) -> list[ModelInfo]:
    """List available Anthropic models.

    Returns:
        List[ModelInfo]: A list of Anthropic model information objects

    """
    anthropic_client = AnthropicFactory.make_anthropic_client(plugin=plugin, backend=backend)
    if not hasattr(anthropic_client, "models"):
        msg = f"{type(anthropic_client).__name__} does not support listing models"
        raise AnthropicSDKUnsupportedError(msg)
    if not isinstance(anthropic_client, AsyncAnthropic):
        msg = "We only support the standard Anthropic client for listing models"
        raise AnthropicSDKUnsupportedError(msg)
    models_response = await anthropic_client.models.list()
    if not models_response:
        msg = "No models found"
        raise AnthropicModelListingError(msg)
    models_list = models_response.data
    if not models_list:
        msg = "No models found"
        raise AnthropicModelListingError(msg)
    return sorted(models_list, key=lambda model: model.created_at)



from pipelex.cogt.extract.extract_input import ExtractInput
from pipelex.cogt.extract.extract_job import ExtractJob
from pipelex.cogt.extract.extract_job_components import ExtractJobConfig, ExtractJobParams, ExtractJobReport
from pipelex.pipeline.job_metadata import JobCategory, JobMetadata


class ExtractJobFactory:
    @classmethod
    def make_extract_job(
        cls,
        extract_input: ExtractInput,
        extract_job_params: ExtractJobParams | None = None,
        extract_job_config: ExtractJobConfig | None = None,
        job_metadata: JobMetadata | None = None,
    ) -> ExtractJob:
        # TODO: manage the param default through the config
        job_metadata = job_metadata or JobMetadata(
            job_category=JobCategory.EXTRACT_JOB,
        )
        job_params = extract_job_params or ExtractJobParams.make_default_extract_job_params()
        job_config = extract_job_config or ExtractJobConfig()
        job_report = ExtractJobReport()

        return ExtractJob(
            job_metadata=job_metadata,
            extract_input=extract_input,
            job_params=job_params,
            job_config=job_config,
            job_report=job_report,
        )



from pydantic import Field, RootModel
from typing_extensions import override

from pipelex.base_exceptions import PipelexError
from pipelex.pipeline.pipeline import Pipeline
from pipelex.pipeline.pipeline_factory import PipelineFactory
from pipelex.pipeline.pipeline_manager_abstract import PipelineManagerAbstract

PipelineManagerRoot = dict[str, Pipeline]


class PipelineManagerNotFoundError(PipelexError):
    pass


class PipelineManager(PipelineManagerAbstract, RootModel[PipelineManagerRoot]):
    root: PipelineManagerRoot = Field(default_factory=dict)

    @override
    def setup(self):
        pass

    @override
    def teardown(self):
        self.root.clear()

    @override
    def get_optional_pipeline(self, pipeline_run_id: str) -> Pipeline | None:
        return self.root.get(pipeline_run_id)

    @override
    def get_pipeline(self, pipeline_run_id: str) -> Pipeline:
        pipeline = self.get_optional_pipeline(pipeline_run_id=pipeline_run_id)
        if pipeline is None:
            msg = f"Pipeline {pipeline_run_id} not found"
            raise PipelineManagerNotFoundError(msg)
        return pipeline

    def _set_pipeline(self, pipeline_run_id: str, pipeline: Pipeline) -> Pipeline:
        self.root[pipeline_run_id] = pipeline
        return pipeline

    @override
    def add_new_pipeline(self) -> Pipeline:
        pipeline = PipelineFactory.make_pipeline()
        self._set_pipeline(pipeline_run_id=pipeline.pipeline_run_id, pipeline=pipeline)
        return pipeline



from pydantic import Field

from pipelex import log
from pipelex.hub import get_secret
from pipelex.system.configuration.config_model import ConfigModel
from pipelex.system.environment import EnvVarNotFoundError, get_required_env
from pipelex.system.exceptions import CredentialsError
from pipelex.tools.secrets.secrets_errors import SecretNotFoundError
from pipelex.types import StrEnum


class AwsCredentialsError(CredentialsError):
    pass


class AwsKeyMethod(StrEnum):
    SECRET_PROVIDER = "secret_provider"
    ENV = "env"


AWS_ACCESS_KEY_ID_VAR_NAME = "AWS_ACCESS_KEY_ID"
AWS_SECRET_ACCESS_KEY_VAR_NAME = "AWS_SECRET_ACCESS_KEY"
AWS_REGION_VAR_NAME = "AWS_REGION"


class AwsConfig(ConfigModel):
    api_key_method: AwsKeyMethod = Field(strict=False)

    def get_aws_access_keys(self) -> tuple[str, str, str]:
        return self.get_aws_access_keys_with_method(api_key_method=self.api_key_method)

    def get_aws_access_keys_with_method(self, api_key_method: AwsKeyMethod) -> tuple[str, str, str]:
        match api_key_method:
            case AwsKeyMethod.ENV:
                log.verbose("Getting AWS access keys from environment (key id and secret access key).")
                try:
                    aws_access_key_id = get_required_env(AWS_ACCESS_KEY_ID_VAR_NAME)
                    aws_secret_access_key = get_required_env(AWS_SECRET_ACCESS_KEY_VAR_NAME)
                    aws_region = get_required_env(AWS_REGION_VAR_NAME)
                except EnvVarNotFoundError as exc:
                    msg = f"Error getting AWS access keys from environment: {exc}"
                    raise AwsCredentialsError(msg) from exc
                log.verbose("Getting AWS region from environment (priority override) or from aws_config.")

            case AwsKeyMethod.SECRET_PROVIDER:
                log.verbose("Getting AWS secret access key from secrets provider (key id and secret access key).")
                try:
                    aws_access_key_id = get_secret(AWS_ACCESS_KEY_ID_VAR_NAME)
                    aws_secret_access_key = get_secret(AWS_SECRET_ACCESS_KEY_VAR_NAME)
                    aws_region = get_secret(AWS_REGION_VAR_NAME)
                except SecretNotFoundError as exc:
                    msg = "Error getting AWS access keys from secrets provider."
                    raise AwsCredentialsError(msg) from exc
                log.verbose("Getting AWS region from environment (priority override) or from aws_config.")

        return aws_access_key_id, aws_secret_access_key, aws_region



from typing_extensions import override

from pipelex.observer.observer_protocol import ObserverProtocol, PayloadKey, PayloadType
from pipelex.system.telemetry.events import EventName, EventProperty, Outcome
from pipelex.system.telemetry.telemetry_manager_abstract import TelemetryManagerAbstract


class ObserverTelemetry(ObserverProtocol):
    def __init__(self, telemetry_manager: TelemetryManagerAbstract):
        self.telemetry_manager = telemetry_manager

    @override
    async def observe_before_run(self, payload: PayloadType) -> None:
        pipe_job = payload[PayloadKey.PIPE_JOB]
        properties = {
            EventProperty.PIPELINE_RUN_ID: payload[PayloadKey.PIPELINE_RUN_ID],
            EventProperty.PIPE_TYPE: pipe_job.pipe_type,
        }
        self.telemetry_manager.track_event(event_name=EventName.PIPE_RUN, properties=properties)

    @override
    async def observe_after_successful_run(self, payload: PayloadType) -> None:
        pipe_job = payload[PayloadKey.PIPE_JOB]
        properties = {
            EventProperty.PIPELINE_RUN_ID: payload[PayloadKey.PIPELINE_RUN_ID],
            EventProperty.PIPE_TYPE: pipe_job.pipe_type,
            EventProperty.PIPE_RUN_OUTCOME: Outcome.SUCCESS,
        }
        self.telemetry_manager.track_event(event_name=EventName.PIPE_COMPLETE, properties=properties)

    @override
    async def observe_after_failing_run(
        self,
        payload: PayloadType,
    ) -> None:
        pipe_job = payload[PayloadKey.PIPE_JOB]
        properties = {
            EventProperty.PIPELINE_RUN_ID: payload[PayloadKey.PIPELINE_RUN_ID],
            EventProperty.PIPE_TYPE: pipe_job.pipe_type,
            EventProperty.PIPE_RUN_OUTCOME: Outcome.FAILURE,
        }
        self.telemetry_manager.track_event(event_name=EventName.PIPE_COMPLETE, properties=properties)



from typing import Literal

from pydantic import Field, field_validator, model_validator
from typing_extensions import override

from pipelex.core.pipes.pipe_blueprint import PipeBlueprint
from pipelex.pipe_controllers.condition.special_outcome import SpecialOutcome
from pipelex.tools.typing.validation_utils import has_exactly_one_among_attributes_from_list
from pipelex.types import Self

OutcomeMap = dict[str, str]


class PipeConditionBlueprint(PipeBlueprint):
    type: Literal["PipeCondition"] = "PipeCondition"
    pipe_category: Literal["PipeController"] = "PipeController"
    expression_template: str | None = None
    expression: str | None = None
    outcomes: OutcomeMap = Field(default_factory=OutcomeMap)
    default_outcome: str | SpecialOutcome
    add_alias_from_expression_to: str | None = None

    @property
    @override
    def pipe_dependencies(self) -> set[str]:
        """Return the set of pipe codes from outcomes and default_pipe_code.

        Excludes special pipe codes like 'continue'.
        """
        pipe_codes = set(self.outcomes.values())
        if self.default_outcome:
            pipe_codes.add(self.default_outcome)
        return pipe_codes - set(SpecialOutcome.value_list())

    @field_validator("outcomes", mode="after")
    @classmethod
    def validate_outcome_map(cls, outcomes: OutcomeMap) -> OutcomeMap:
        if not outcomes:
            msg = f"PipeConditionBlueprint must have at least one mapping in outcomes, got: {outcomes}"
            raise ValueError(msg)
        return outcomes

    @model_validator(mode="after")
    def validate_expression_and_expression_template(self) -> Self:
        if not has_exactly_one_among_attributes_from_list(self, attributes_list=["expression_template", "expression"]):
            msg = "PipeCondition should have exactly one of 'expression_template' or 'expression'"
            raise ValueError(msg)
        return self

    @override
    def validate_inputs(self):
        pass

    @override
    def validate_output(self):
        pass



import json

from typing_extensions import override
from yattag import Doc

from pipelex.core.stuffs.stuff_content import StuffContent


class MermaidContent(StuffContent):
    mermaid_code: str
    mermaid_url: str

    @property
    @override
    def short_desc(self) -> str:
        return f"some mermaid code ({len(self.mermaid_code)} chars)"

    @override
    def __str__(self) -> str:
        return self.mermaid_code

    @override
    def rendered_plain(self) -> str:
        return self.mermaid_code

    @override
    def rendered_html(self) -> str:
        doc, tag, text = Doc().tagtext()
        with tag("div", klass="mermaid"):
            text(self.mermaid_code)
        return doc.getvalue()

    @override
    def rendered_markdown(self, level: int = 1, is_pretty: bool = False) -> str:
        return self.mermaid_code

    @override
    def rendered_json(self) -> str:
        return json.dumps({"mermaid": self.mermaid_code})



from pipelex.core.pipes.variable_multiplicity import make_variable_multiplicity
from pipelex.pipe_controllers.sub_pipe import SubPipe
from pipelex.pipe_controllers.sub_pipe_blueprint import SubPipeBlueprint
from pipelex.pipe_run.pipe_run_params import BatchParams


class SubPipeFactory:
    @classmethod
    def make_from_blueprint(
        cls,
        blueprint: SubPipeBlueprint,
    ) -> SubPipe:
        """Create a SubPipe from a SubPipeBlueprint."""
        output_multiplicity = make_variable_multiplicity(
            nb_items=blueprint.nb_output,
            multiple_items=blueprint.multiple_output,
        )
        batch_params: BatchParams | None = None
        if blueprint.batch_over and blueprint.batch_as:
            batch_params = BatchParams.make_batch_params(
                input_list_name=blueprint.batch_over,
                input_item_name=blueprint.batch_as,
            )
        else:
            batch_params = None
        return SubPipe(
            pipe_code=blueprint.pipe,
            output_name=blueprint.result,
            output_multiplicity=output_multiplicity,
            batch_params=batch_params,
        )



from collections.abc import Callable, Sequence
from typing import Any, TypeVar, cast

from pydantic import BaseModel, ValidationError
from rich.repr import Result as RichReprResult
from typing_extensions import override

from pipelex.tools.misc.attribute_utils import AttributePolisher
from pipelex.types import StrEnum

BaseModelTypeVar = TypeVar("BaseModelTypeVar", bound=BaseModel)

T = TypeVar("T")


def empty_list_factory_of(_: type[T]) -> Callable[[], list[T]]:
    def _factory() -> list[T]:
        return []

    return _factory


class PydanticValidationErrorAnalysis(BaseModel):
    error_msg: str

    missing_fields: list[str]
    extra_fields: list[str]
    type_errors: list[str]
    value_errors: list[str]
    enum_errors: list[str]
    union_tag_errors: list[str]
    model_type_errors: list[str]


def analyze_pydantic_validation_error(exc: ValidationError) -> PydanticValidationErrorAnalysis:
    """Analyze a Pydantic ValidationError into a readable string with detailed error information.

    Args:
        exc: The Pydantic ValidationError exception

    Returns:
        A PydanticValidationErrorAnalysis object containing categorized validation errors

    """
    error_msg = "Validation error(s):"

    # Collect different types of validation errors
    missing_fields = [f"'{'.'.join(map(str, err['loc']))}'" for err in exc.errors() if err["type"] == "missing"]
    extra_fields = [f"'{'.'.join(map(str, err['loc']))}'" for err in exc.errors() if err["type"] == "extra_forbidden"]
    type_errors = [f"'{'.'.join(map(str, err['loc']))}': expected {err['type']}" for err in exc.errors() if err["type"] == "type_error"]
    value_errors = [f"'{'.'.join(map(str, err['loc']))}': {err['msg']}" for err in exc.errors() if err["type"] == "value_error"]
    enum_errors = [
        f"'{'.'.join(map(str, err['loc']))}': invalid enum value '{err.get('input', 'unknown')}'" for err in exc.errors() if err["type"] == "enum"
    ]
    union_tag_errors: list[str] = []
    for err in exc.errors():
        if err["type"] == "union_tag_not_found":
            field_path = ".".join(map(str, err["loc"]))
            # Extract discriminator field name from context
            discriminator = err.get("ctx", {}).get("discriminator", "type")
            union_tag_errors.append(f"'{field_path}': missing required discriminator field '{discriminator}'")

    model_type_errors: list[str] = []
    for err in exc.errors():
        if err["type"] == "model_type":
            field_path = ".".join(map(str, err["loc"]))
            # Extract expected type from context if available
            expected_type = err.get("ctx", {}).get("class_name", "unknown model type")
            actual_input = err.get("input", "unknown")
            actual_type = type(actual_input).__name__ if actual_input != "unknown" else "unknown"
            model_type_errors.append(f"'{field_path}': expected {expected_type}, got {actual_type}")

    # Add each type of error to the message if present
    if missing_fields:
        error_msg += f"\n\nMissing required fields: {', '.join(missing_fields)}"
    if extra_fields:
        error_msg += f"\n\nExtra forbidden fields: {', '.join(extra_fields)}"
    if type_errors:
        error_msg += f"\n\nType errors: {', '.join(type_errors)}"
    if value_errors:
        error_msg += f"\n\nValue errors: {', '.join(value_errors)}"
    if enum_errors:
        error_msg += f"\n\nEnum errors: {', '.join(enum_errors)}"
    if union_tag_errors:
        error_msg += f"\n\nUnion discriminator errors: {', '.join(union_tag_errors)}"
    if model_type_errors:
        error_msg += f"\n\nModel type errors: {', '.join(model_type_errors)}"

    # If none of the specific error types were found, add the raw error messages
    if not any([missing_fields, extra_fields, type_errors, value_errors, enum_errors, union_tag_errors, model_type_errors]):
        error_msg += "\n\nOther validation errors:"
        for err in exc.errors():
            error_msg += f"\n{'.'.join(map(str, err['loc']))}: {err['type']}: {err['msg']}"

    return PydanticValidationErrorAnalysis(
        error_msg=error_msg,
        missing_fields=missing_fields,
        extra_fields=extra_fields,
        type_errors=type_errors,
        value_errors=value_errors,
        enum_errors=enum_errors,
        union_tag_errors=union_tag_errors,
        model_type_errors=model_type_errors,
    )


def format_pydantic_validation_error(exc: ValidationError) -> str:
    """Format a Pydantic ValidationError into a readable string with detailed error information.

    Args:
        exc: The Pydantic ValidationError exception

    Returns:
        A formatted string containing categorized validation errors

    """
    return analyze_pydantic_validation_error(exc).error_msg


def convert_strenum_to_str(
    obj: dict[str, Any] | list[Any] | StrEnum | Any,
) -> dict[str, Any] | list[Any] | str | Any:
    if isinstance(obj, dict):
        obj_dict = cast("dict[str, Any]", obj)
        return {str(key): convert_strenum_to_str(value) for key, value in obj_dict.items()}
    elif isinstance(obj, list):
        obj_list = cast("list[Any]", obj)
        return [convert_strenum_to_str(item) for item in obj_list]
    elif isinstance(obj, StrEnum):
        if hasattr(obj, "display_name"):
            return obj.display_name()  # pyright: ignore[reportUnknownVariableType, reportUnknownMemberType, reportAttributeAccessIssue]
        else:
            return str(obj)
    else:
        return obj


class ExtraFieldAttribute(StrEnum):
    IS_HIDDEN = "is_hidden"


class FieldVisibility(StrEnum):
    ALL_FIELDS = "all_fields"
    NO_HIDDEN_FIELDS = "no_hidden_fields"
    ONLY_HIDDEN_FIELDS = "only_hidden_fields"


def clean_model_to_dict(obj: BaseModel) -> dict[str, Any]:
    dict_dump = serialize_model(
        obj=obj,
        field_visibility=FieldVisibility.NO_HIDDEN_FIELDS,
        is_stringify_enums=True,
    )
    if not isinstance(dict_dump, dict):
        msg = f"Expected dict, got {type(dict_dump)}"
        raise TypeError(msg)
    return cast("dict[str, Any]", dict_dump)


def serialize_model(
    obj: Any,
    field_visibility: FieldVisibility = FieldVisibility.NO_HIDDEN_FIELDS,
    is_stringify_enums: bool = True,
) -> dict[str, Any] | list[Any] | Any:
    """Recursively serialize a Pydantic BaseModel (and its nested BaseModels)
    into a dictionary, omitting any fields marked with
    'json_schema_extra={ExtraFieldAttribute.IS_HIDDEN: True}'.

    If 'obj' is not a BaseModel, return it as-is (useful for nested lists/dicts).
    """
    # If it's not a Pydantic model, return it directly
    if not isinstance(obj, BaseModel):
        # Might be a primitive type, list, dict, etc.
        # We only handle nesting if it's inside BaseModels
        return obj

    # Identify which fields should be excluded
    fields_to_exclude: set[str] = set()

    for field_name, field_info in obj.__class__.model_fields.items():
        json_schema_extra = field_info.json_schema_extra
        is_hidden: bool
        if json_schema_extra and isinstance(json_schema_extra, dict):
            typed_json_schema_extra = cast("dict[str, Any]", json_schema_extra)
            is_hidden = typed_json_schema_extra.get(ExtraFieldAttribute.IS_HIDDEN) is True
        else:
            is_hidden = False
        match field_visibility:
            case FieldVisibility.ALL_FIELDS:
                pass
            case FieldVisibility.NO_HIDDEN_FIELDS:
                if is_hidden:
                    fields_to_exclude.add(field_name)
            case FieldVisibility.ONLY_HIDDEN_FIELDS:
                if not is_hidden:
                    fields_to_exclude.add(field_name)

    # Build a dict, omitting hidden fields. Recursively handle nested models.
    data: dict[str, Any] = {}
    for field_name in obj.__class__.model_fields:
        if field_name in fields_to_exclude:
            continue  # Skip hidden fields

        value = getattr(obj, field_name)

        # If the value is another BaseModel, recurse
        if isinstance(value, BaseModel):
            data[field_name] = serialize_model(
                obj=value,
                field_visibility=field_visibility,
                is_stringify_enums=is_stringify_enums,
            )

        # If it's a list, we recurse for each item
        elif isinstance(value, list):
            value_list = cast("list[Any]", value)
            data[field_name] = [
                serialize_model(
                    obj=item,
                    field_visibility=field_visibility,
                    is_stringify_enums=is_stringify_enums,
                )
                for item in value_list
            ]

        # If it's a dict, we can similarly recurse for any nested BaseModels inside the dict
        elif isinstance(value, dict):
            value_dict = cast("dict[str, Any]", value)
            data[field_name] = {
                key: serialize_model(
                    obj=value,
                    field_visibility=field_visibility,
                    is_stringify_enums=is_stringify_enums,
                )
                for key, value in value_dict.items()
            }

        elif is_stringify_enums and isinstance(value, StrEnum):
            if hasattr(value, "display_name"):
                data[field_name] = value.display_name()  # pyright: ignore[reportUnknownMemberType, reportAttributeAccessIssue]
            else:
                data[field_name] = str(value)

        # Otherwise, just store the raw value
        else:
            data[field_name] = value

    return data


class CustomBaseModel(BaseModel):
    @override
    def __rich_repr__(self) -> RichReprResult:
        for item in super().__rich_repr__():  # type: ignore[misc]
            if isinstance(item, tuple):
                tuple_item = cast("tuple[Any, ...]", item)
                if len(tuple_item) >= 2:
                    name = tuple_item[0]
                    value = tuple_item[1]
                    if AttributePolisher.should_truncate(name=name, value=value):
                        truncated_value = AttributePolisher.get_truncated_value(name, value)
                        if len(tuple_item) == 3:
                            yield name, truncated_value, tuple_item[2]
                        else:
                            yield name, truncated_value
                    else:
                        yield item
            else:
                yield item

    @override
    def __repr_args__(self) -> Sequence[tuple[str | None, Any]]:
        processed_args: list[tuple[str | None, Any]] = []
        for name, value in super().__repr_args__():
            if name and AttributePolisher.should_truncate(name=name, value=value):
                truncated_value = AttributePolisher.get_truncated_value(name, value)
                processed_args.append((name, truncated_value))
            else:
                processed_args.append((name, value))
        return processed_args

    def model_dump_truncated(self, **kwargs: Any) -> Any:
        """Dump the model to a dictionary with serialize_as_any=True and apply

        AttributePolisher truncation to fields that should be truncated.
        Handles nested attributes recursively.

        Args:
            **kwargs: Additional keyword arguments to pass to model_dump

        Returns:
            Dictionary with truncated values where appropriate

        """
        # Get the model dump with serialize_as_any=True
        dumped_data = self.model_dump(**kwargs)

        # Apply truncation logic recursively
        return self._apply_truncation_recursive(dumped_data)

    def _apply_truncation_recursive(self, obj: Any, name: str | None = None) -> Any:
        """Recursively apply AttributePolisher truncation logic to a data structure.

        Args:
            obj: The object to process
            name: The field name (for truncation logic)

        Returns:
            The processed object with truncation applied where appropriate

        """
        # First check if this specific object should be truncated
        if name and AttributePolisher.should_truncate(name=name, value=obj):
            return AttributePolisher.get_truncated_value(name, obj)

        # If it's a dictionary, recurse into its values
        if isinstance(obj, dict):
            obj_dict = cast("dict[str, Any]", obj)
            truncated_dict: dict[str, Any] = {}
            for key, value in obj_dict.items():
                truncated_dict[key] = self._apply_truncation_recursive(value, name=key)
            return truncated_dict

        # If it's a list, recurse into its items
        if isinstance(obj, list):
            obj_list = cast("list[Any]", obj)
            return [self._apply_truncation_recursive(item, name=name) for item in obj_list]

        # If it's a tuple, recurse into its items and return as tuple
        if isinstance(obj, tuple):
            return tuple(self._apply_truncation_recursive(item, name=name) for item in obj)  # pyright: ignore[reportUnknownVariableType]

        # For all other types, return as-is
        return obj



from pipelex.plugins.plugin_sdk_registry import PluginSdkRegistry


class PluginManager:
    def __init__(self):
        self.plugin_sdk_registry = PluginSdkRegistry()

    def setup(self):
        pass

    def teardown(self):
        self.plugin_sdk_registry.teardown()



from typing import Any

from pydantic import BaseModel, Field, model_validator

from pipelex.cogt.templating.template_category import TemplateCategory
from pipelex.cogt.templating.templating_style import TemplatingStyle
from pipelex.tools.jinja2.jinja2_errors import Jinja2TemplateSyntaxError
from pipelex.tools.jinja2.jinja2_parsing import check_jinja2_parsing


class TemplateBlueprint(BaseModel):
    template: str = Field(description="Raw template source")
    templating_style: TemplatingStyle | None = Field(default=None, description="Style of prompting to use (typically for different LLMs)")
    category: TemplateCategory = Field(
        description="Category of the template (could also be HTML, MARKDOWN, MERMAID, etc.), influences template rendering rules",
    )
    extra_context: dict[str, Any] | None = Field(default=None, description="Additional context variables for template rendering")

    @model_validator(mode="after")
    def validate_template(self) -> "TemplateBlueprint":
        try:
            check_jinja2_parsing(template_source=self.template, template_category=self.category)
        except Jinja2TemplateSyntaxError as exc:
            msg = f"Could not parse template for TemplateBlueprint: {exc}"
            raise ValueError(msg) from exc
        return self



from abc import ABC, abstractmethod
from typing import Any

from jinja2.runtime import Context

from pipelex.types import StrEnum


class Jinja2FilterName(StrEnum):
    FORMAT = "format"
    TAG = "tag"


class Jinja2ContextKey(StrEnum):
    TAG_STYLE = "tag_style"
    TEXT_FORMAT = "text_format"


class Jinja2TaggableAbstract(ABC):
    @abstractmethod
    def render_tagged_for_jinja2(self, context: Context, tag_name: str | None = None) -> tuple[Any, str | None]:
        pass



from typing import Annotated

import typer
from click import Command, Context
from typer.core import TyperGroup
from typing_extensions import override

from pipelex.cli.commands.build_cmd import build_app
from pipelex.cli.commands.doctor_cmd import doctor_cmd
from pipelex.cli.commands.init.command import init_cmd
from pipelex.cli.commands.init.ui.types import InitFocus
from pipelex.cli.commands.kit_cmd import kit_app
from pipelex.cli.commands.run_cmd import run_cmd
from pipelex.cli.commands.show_cmd import show_app
from pipelex.cli.commands.validate_cmd import validate_cmd
from pipelex.cli.readiness import check_readiness
from pipelex.hub import get_console
from pipelex.tools.misc.package_utils import get_package_version


class PipelexCLI(TyperGroup):
    @override
    def list_commands(self, ctx: Context) -> list[str]:
        # List the commands in the proper order because natural ordering doesn't work between Typer groups and commands
        return ["init", "doctor", "kit", "build", "validate", "run", "show"]

    @override
    def get_command(self, ctx: Context, cmd_name: str) -> Command | None:
        cmd = super().get_command(ctx, cmd_name)
        if cmd is None:
            typer.echo(f"Unknown command: {cmd_name}")
            typer.echo(ctx.get_help())
            ctx.exit(1)
        return cmd


def main() -> None:
    """Entry point for the pipelex CLI."""
    app()


app = typer.Typer(
    add_completion=False,
    no_args_is_help=True,
    pretty_exceptions_show_locals=False,
    cls=PipelexCLI,
)


@app.callback(invoke_without_command=True)
def app_callback(ctx: typer.Context) -> None:
    console = get_console()
    package_version = get_package_version()
    console.print(
        f"""

░█████████  ░[bold green4]██[/bold green4]                      ░██
░██     ░██                          ░██
░██     ░██ ░██░████████   ░███████  ░██  ░███████  ░██    ░[bold green4]██[/bold green4]
░█████████  ░██░██    ░██ ░██    ░██ ░██ ░██    ░██  ░██  ░██
░██         ░██░██    ░██ ░█████████ ░██ ░█████████   ░█████
░██         ░██░███   ░██ ░██        ░██ ░██         ░██  ░██
░██         ░██░██░█████   ░███████  ░██  ░███████  ░██    ░██
               ░██
               ░██                                     v[cyan]{package_version}[/cyan]
"""
    )
    # Skip checks if no command is being run (e.g., just --help) or if running init/doctor command
    if ctx.invoked_subcommand is None or ctx.invoked_subcommand in ("init", "doctor"):
        return

    # Check system readiness (dependencies and venv for dev installs)
    check_readiness()

    init_cmd(silent=True)


@app.command(name="init", help="Initialize Pipelex configuration in a `.pipelex` directory")
def init_command(
    focus: Annotated[InitFocus, typer.Argument(help="What to initialize: 'config', 'telemetry', or 'all'")] = InitFocus.ALL,
    reset: Annotated[bool, typer.Option("--reset", "-r", help="Reset existing configuration files")] = False,
) -> None:
    """Initialize Pipelex configuration and telemetry."""
    init_cmd(focus=focus, reset=reset)


@app.command(name="doctor", help="Check Pipelex configuration health and suggest fixes")
def doctor_command(
    fix: Annotated[bool, typer.Option("--fix", "-f", help="Offer to fix detected issues interactively")] = False,
) -> None:
    """Check Pipelex configuration health."""
    doctor_cmd(fix=fix)


app.add_typer(kit_app, name="kit", help="Manage kit assets: agent rules, migration rules")
app.add_typer(
    build_app, name="build", help="Generate AI workflows from natural language requirements: pipelines in .plx format and python code to run them"
)
app.command(name="validate", help="Validate pipes: static validation for syntax and dependencies, dry-run execution for logic and consistency")(
    validate_cmd
)
app.command(name="run", help="Run a pipe, optionally providing a specific bundle file (.plx)")(run_cmd)
app.add_typer(show_app, name="show", help="Show configuration, pipes, and list AI models")



from typing import TYPE_CHECKING, Literal

from pydantic import Field, field_validator
from pydantic.json_schema import SkipJsonSchema
from rich.console import Group
from rich.text import Text
from typing_extensions import override

from pipelex.builder.pipe.pipe_spec import PipeSpec
from pipelex.pipe_operators.extract.pipe_extract_blueprint import PipeExtractBlueprint
from pipelex.tools.misc.pretty import PrettyPrintable
from pipelex.types import StrEnum

if TYPE_CHECKING:
    from pipelex.cogt.extract.extract_setting import ExtractModelChoice


class ExtractSkill(StrEnum):
    EXTRACT_TEXT_FROM_VISUALS = "extract_text_from_visuals"
    EXTRACT_TEXT_FROM_PDF = "extract_text_from_pdf"


class PipeExtractSpec(PipeSpec):
    """Spec for OCR (Optical Character Recognition) pipe operations in the Pipelex framework.

    PipeExtract enables text extraction from images and documents using OCR technology.
    Supports various OCR platforms and output configurations including image detection,
    caption generation, and page rendering.

    Validation Rules:
        - inputs dict must have exactly one input entry, and the value must be either `Image` or `PDF`.
        - output must be "Page"
    """

    type: SkipJsonSchema[Literal["PipeExtract"]] = "PipeExtract"
    pipe_category: SkipJsonSchema[Literal["PipeOperator"]] = "PipeOperator"
    extract_skill: ExtractSkill | str = Field(description="Select the most adequate extraction model skill according to the task to be performed.")
    page_images: bool | None = Field(default=None, description="Whether to include detected images in the Extract output.")
    page_image_captions: bool | None = Field(default=None, description="Whether to generate captions for detected images using AI.")
    page_views: bool | None = Field(default=None, description="Whether to include rendered page views in the output.")

    @override
    @field_validator("output", mode="before")
    @classmethod
    def validate_output(cls, output: str) -> str:
        return "Page[]"

    @field_validator("extract_skill", mode="before")
    @classmethod
    def validate_extract_skill(cls, extract_skill_value: str) -> ExtractSkill:
        return ExtractSkill(extract_skill_value)

    @field_validator("inputs", mode="before")
    @classmethod
    def validate_extract_inputs(cls, inputs_value: dict[str, str] | None) -> dict[str, str] | None:
        if inputs_value is None:
            msg = "PipeExtract must have exactly one input which must be either `Image` or `PDF`."
            raise ValueError(msg)
        if len(inputs_value) != 1:
            msg = "PipeExtract must have exactly one input which must be either `Image` or `PDF`."
            raise ValueError(msg)
        return inputs_value

    @override
    def rendered_pretty(self, title: str | None = None, depth: int = 0) -> PrettyPrintable:
        # Get base pipe information from parent
        base_group = super().rendered_pretty(title=title, depth=depth)

        # Create a group combining base info with extract-specific details
        extract_group = Group()
        extract_group.renderables.append(base_group)

        # Add extract specific information
        extract_group.renderables.append(Text())  # Blank line
        extract_group.renderables.append(Text.from_markup(f"Extract Skill: [bold yellow]{self.extract_skill}[/bold yellow]"))

        # Add optional extraction settings if they are set
        if self.page_images is not None:
            extract_group.renderables.append(Text.from_markup(f"Include Page Images: [bold magenta]{self.page_images}[/bold magenta]"))
        if self.page_image_captions is not None:
            extract_group.renderables.append(Text.from_markup(f"Generate Image Captions: [bold magenta]{self.page_image_captions}[/bold magenta]"))
        if self.page_views is not None:
            extract_group.renderables.append(Text.from_markup(f"Include Page Views: [bold magenta]{self.page_views}[/bold magenta]"))

        return extract_group

    @override
    def to_blueprint(self) -> PipeExtractBlueprint:
        base_blueprint = super().to_blueprint()

        # create extract choice as a str
        extract_model_choice: ExtractModelChoice = self.extract_skill

        return PipeExtractBlueprint(
            source=None,
            description=base_blueprint.description,
            inputs=base_blueprint.inputs,
            output=base_blueprint.output,
            model=extract_model_choice,
            page_images=self.page_images,
            page_image_captions=self.page_image_captions,
            page_views=self.page_views,
            page_views_dpi=None,
        )



from __future__ import annotations

import asyncio
import pathlib
from typing import TYPE_CHECKING

import pypdfium2 as pdfium
from pypdfium2.raw import FPDFBitmap_BGRA

from pipelex.system.exceptions import ToolError
from pipelex.tools.misc.file_fetch_utils import fetch_file_from_url_httpx_async
from pipelex.tools.misc.path_utils import clarify_path_or_url

if TYPE_CHECKING:
    from PIL import Image
PDFIUM2_REFERENCE_DPI = 72


class PyPdfium2RendererError(ToolError):
    pass


PdfInput = str | pathlib.Path | bytes


class PyPdfium2Renderer:
    """Thread-safe PDF page renderer built on pypdfium2.

    • All entry into the native PDFium library is protected by a single
      asyncio.Lock, so the enclosing *process* is safe even if other
      libraries spin up worker threads.

    • Heavy work runs inside `asyncio.to_thread`, keeping the event-loop
      responsive for the rest of your application.
    """

    _pdfium_lock: asyncio.Lock = asyncio.Lock()  # shared per process

    # ---- internal blocking helper ------------------------------------
    # TODO: Needs UT
    @staticmethod
    def _render_pdf_pages_sync(pdf_input: PdfInput, scale: float) -> list[Image.Image]:
        pdf_doc = pdfium.PdfDocument(pdf_input)
        images: list[Image.Image] = []
        for index in range(len(pdf_doc)):
            # page = doc.get_page(index)
            page = pdf_doc[index]

            pil_img: Image.Image = page.render(  # pyright: ignore[reportUnknownMemberType, reportUnknownVariableType]
                scale=scale,  # pyright: ignore[reportArgumentType]
                force_bitmap_format=FPDFBitmap_BGRA,  # always 4-channel
                rev_byteorder=True,  # so we get RGBA
            ).to_pil()

            images.append(pil_img)  # pyright: ignore[reportUnknownArgumentType]
            page.close()
        pdf_doc.close()
        return images

    # TODO: Needs UT
    @staticmethod
    def _get_text_from_pdf_pages_sync(pdf_input: PdfInput) -> list[str]:
        pdf_doc = pdfium.PdfDocument(pdf_input)
        texts: list[str] = []
        for index in range(len(pdf_doc)):
            page = pdf_doc[index]
            text = page.get_textpage().get_text_bounded()  # pyright: ignore[reportUnknownMemberType]
            texts.append(text)  # pyright: ignore[reportUnknownArgumentType]
            page.close()
        pdf_doc.close()
        return texts

    # ---- public async façade -----------------------------------------
    async def render_pdf_pages(self, pdf_input: PdfInput, dpi: int) -> list[Image.Image]:
        scale = dpi / PDFIUM2_REFERENCE_DPI
        """Render *one* page and return PNG bytes."""
        async with self._pdfium_lock:
            return await asyncio.to_thread(self._render_pdf_pages_sync, pdf_input, scale)

    async def get_text_from_pdf_pages(self, pdf_input: PdfInput) -> list[str]:
        """Extract text from all pages of a PDF."""
        async with self._pdfium_lock:
            return await asyncio.to_thread(self._get_text_from_pdf_pages_sync, pdf_input)

    async def render_pdf_pages_from_uri(self, pdf_uri: str, dpi: int) -> list[Image.Image]:
        pdf_path, pdf_url = clarify_path_or_url(path_or_uri=pdf_uri)
        if pdf_url:
            pdf_bytes = await fetch_file_from_url_httpx_async(url=pdf_url)
            return await self.render_pdf_pages(pdf_input=pdf_bytes, dpi=dpi)
        if pdf_path:
            return await self.render_pdf_pages(pdf_input=pdf_path, dpi=dpi)
        msg = f"Invalid PDF URI: {pdf_uri}"
        raise PyPdfium2RendererError(msg)

    async def get_text_from_pdf_pages_from_uri(self, pdf_uri: str) -> list[str]:
        """Extract text from all pages of a PDF from URI."""
        pdf_path, pdf_url = clarify_path_or_url(path_or_uri=pdf_uri)
        if pdf_url:
            pdf_bytes = await fetch_file_from_url_httpx_async(url=pdf_url)
            return await self.get_text_from_pdf_pages(pdf_input=pdf_bytes)
        if pdf_path:
            return await self.get_text_from_pdf_pages(pdf_input=pdf_path)
        msg = f"Invalid PDF URI: {pdf_uri}"
        raise PyPdfium2RendererError(msg)


pypdfium2_renderer = PyPdfium2Renderer()



"""General UI components for the init command."""

from rich.console import Console
from rich.panel import Panel
from rich.prompt import Confirm

from pipelex.cli.commands.init.ui.types import InitFocus


def display_already_configured_message(focus: InitFocus, console: Console, config_path: str) -> bool:
    """Display 'already configured' message and ask if user wants to reconfigure.

    Args:
        focus: The initialization focus area.
        console: Rich Console instance for output.
        config_path: Path to the configuration file.

    Returns:
        True if user wants to reconfigure, False otherwise.
    """
    # Mapping of focus to (subject, action_verb)
    focus_messages = {
        InitFocus.INFERENCE: ("Inference backends", "inference backends"),
        InitFocus.ROUTING: ("Routing profile", "routing profile"),
        InitFocus.TELEMETRY: ("Telemetry preferences", "telemetry preferences"),
        InitFocus.CONFIG: ("Configuration files", "configuration"),
    }

    if focus == InitFocus.ALL:
        console.print()
        console.print("[green]✓[/green] Pipelex is already fully initialized!")
        console.print()
        console.print("[dim]Configuration files are in place:[/dim] [cyan].pipelex/[/cyan]")
        console.print("[dim]Telemetry preferences are configured[/dim]")
        console.print()
        console.print("[dim]💡 Tip: Use[/dim] [cyan]--reset[/cyan] [dim]to reconfigure or troubleshoot:[/dim]")
        console.print("   [cyan]pipelex init --reset[/cyan]")
        console.print()
        return False

    if focus == InitFocus.CONFIG:
        console.print()
        console.print("[green]✓[/green] Configuration files are already in place!")
        console.print()
        console.print("[dim]Configuration directory:[/dim] [cyan].pipelex/[/cyan]")
        console.print()
        console.print("[dim]💡 Tip: Use[/dim] [cyan]--reset[/cyan] [dim]to reconfigure or troubleshoot:[/dim]")
        console.print(f"   [cyan]pipelex init {focus} --reset[/cyan]")
        console.print()
        return False

    if focus in focus_messages:
        subject, action_verb = focus_messages[focus]
        console.print()
        if focus == InitFocus.ROUTING:
            console.print(f"[green]✓[/green] {subject} is already configured!")
        else:
            console.print(f"[green]✓[/green] {subject} are already configured!")
        console.print()
        console.print(f"[dim]Configuration file:[/dim] [cyan]{config_path}[/cyan]")
        console.print()

        return Confirm.ask(f"[bold]Would you like to reconfigure {action_verb}?[/bold]", default=False)

    return False


def build_initialization_panel(needs_config: bool, needs_inference: bool, needs_routing: bool, needs_telemetry: bool, reset: bool) -> Panel:
    """Build the initialization confirmation panel.

    Args:
        needs_config: Whether config initialization is needed.
        needs_inference: Whether inference setup is needed.
        needs_routing: Whether routing setup is needed.
        needs_telemetry: Whether telemetry setup is needed.
        reset: Whether this is a reset operation.

    Returns:
        A Panel containing the initialization confirmation message.
    """
    # Build message based on what's being initialized
    message_parts: list[str] = []
    if reset:
        if needs_config:
            message_parts.append("• [yellow]Reset and reconfigure[/yellow] configuration files in [cyan].pipelex/[/cyan]")
        if needs_inference:
            message_parts.append("• [yellow]Reset and reconfigure[/yellow] inference backends")
        if needs_routing:
            message_parts.append("• [yellow]Reset and reconfigure[/yellow] routing profile")
        if needs_telemetry:
            message_parts.append("• [yellow]Reset and reconfigure[/yellow] telemetry preferences")
    else:
        if needs_config:
            message_parts.append("• Create required configuration files in [cyan].pipelex/[/cyan]")
        if needs_inference:
            message_parts.append("• Ask you to choose your inference backends")
        if needs_routing:
            message_parts.append("• Ask you to configure your routing profile")
        if needs_telemetry:
            message_parts.append("• Ask you to choose your telemetry preferences")

    # Determine title based on what's being initialized
    num_items = sum([needs_config, needs_inference, needs_routing, needs_telemetry])
    if reset:
        if num_items > 1:
            title_text = "[bold yellow]Resetting Configuration[/bold yellow]"
        elif needs_config:
            title_text = "[bold yellow]Resetting Configuration Files[/bold yellow]"
        elif needs_inference:
            title_text = "[bold yellow]Resetting Inference Backends[/bold yellow]"
        elif needs_routing:
            title_text = "[bold yellow]Resetting Routing Profile[/bold yellow]"
        else:
            title_text = "[bold yellow]Resetting Telemetry[/bold yellow]"
    elif num_items > 1:
        title_text = "[bold cyan]Pipelex Initialization[/bold cyan]"
    elif needs_config:
        title_text = "[bold cyan]Configuration Setup[/bold cyan]"
    elif needs_inference:
        title_text = "[bold cyan]Inference Backend Setup[/bold cyan]"
    elif needs_routing:
        title_text = "[bold cyan]Routing Profile Setup[/bold cyan]"
    else:
        title_text = "[bold cyan]Telemetry Setup[/bold cyan]"

    message = "\n".join(message_parts)
    border_color = "yellow" if reset else "cyan"

    return Panel(
        message,
        title=title_text,
        border_style=border_color,
        padding=(1, 2),
    )



import importlib.util

from pipelex.cogt.exceptions import MissingDependencyError
from pipelex.cogt.extract.extract_worker_abstract import ExtractWorkerAbstract
from pipelex.cogt.model_backends.model_spec import InferenceModelSpec
from pipelex.hub import get_models_manager, get_plugin_manager
from pipelex.plugins.plugin_sdk_registry import Plugin
from pipelex.reporting.reporting_protocol import ReportingProtocol


class ExtractWorkerFactory:
    def make_extract_worker(
        self,
        inference_model: InferenceModelSpec,
        reporting_delegate: ReportingProtocol | None = None,
    ) -> ExtractWorkerAbstract:
        plugin = Plugin.make_for_inference_model(inference_model=inference_model)
        backend = get_models_manager().get_required_inference_backend(inference_model.backend_name)
        plugin_sdk_registry = get_plugin_manager().plugin_sdk_registry
        extract_worker: ExtractWorkerAbstract
        match plugin.sdk:
            case "mistral":
                if importlib.util.find_spec("mistralai") is None:
                    lib_name = "mistralai"
                    lib_extra_name = "mistral"
                    msg = "The mistralai SDK is required in order to use Mistral OCR models through the mistralai client."
                    raise MissingDependencyError(
                        lib_name,
                        lib_extra_name,
                        msg,
                    )

                from pipelex.plugins.mistral.mistral_extract_worker import MistralExtractWorker  # noqa: PLC0415
                from pipelex.plugins.mistral.mistral_factory import MistralFactory  # noqa: PLC0415

                extract_sdk_instance = plugin_sdk_registry.get_sdk_instance(plugin=plugin) or plugin_sdk_registry.set_sdk_instance(
                    plugin=plugin,
                    sdk_instance=MistralFactory.make_mistral_client(backend=backend),
                )

                extract_worker = MistralExtractWorker(
                    sdk_instance=extract_sdk_instance,
                    extra_config=backend.extra_config,
                    inference_model=inference_model,
                    reporting_delegate=reporting_delegate,
                )
            case "pypdfium2":
                from pipelex.plugins.pypdfium2.pypdfium2_worker import Pypdfium2Worker  # noqa: PLC0415

                extract_worker = Pypdfium2Worker(
                    extra_config=backend.extra_config,
                    inference_model=inference_model,
                    reporting_delegate=reporting_delegate,
                )
            case _:
                msg = f"Plugin '{plugin}' is not supported"
                raise NotImplementedError(msg)

        return extract_worker



from typing import Any

from typing_extensions import override

from pipelex.core.concepts.concept import Concept
from pipelex.core.pipes.inputs.input_requirements import InputRequirements
from pipelex.core.pipes.pipe_factory import PipeFactoryProtocol
from pipelex.pipe_controllers.condition.exceptions import PipeConditionFactoryError
from pipelex.pipe_controllers.condition.pipe_condition import PipeCondition
from pipelex.pipe_controllers.condition.pipe_condition_blueprint import PipeConditionBlueprint


class PipeConditionFactory(PipeFactoryProtocol[PipeConditionBlueprint, PipeCondition]):
    @classmethod
    @override
    def make(
        cls,
        pipe_category: Any,
        pipe_type: str,
        pipe_code: str,
        domain_code: str,
        description: str | None,
        inputs: InputRequirements,
        output: Concept,
        blueprint: PipeConditionBlueprint,
    ) -> PipeCondition:
        # Compute expression from expression_template or expression in blueprint
        expression: str | None = None
        if blueprint.expression_template:
            expression = blueprint.expression_template
        elif blueprint.expression:
            expression = "{{ " + blueprint.expression + " }}"
        else:
            msg = "PipeCondition must have either expression_template or expression"
            raise PipeConditionFactoryError(msg)

        return PipeCondition(
            domain=domain_code,
            code=pipe_code,
            description=description,
            inputs=inputs,
            output=output,
            expression=expression,
            outcome_map=blueprint.outcomes,
            default_outcome=blueprint.default_outcome,
            add_alias_from_expression_to=blueprint.add_alias_from_expression_to,
        )



from pipelex.base_exceptions import PipelexError
from pipelex.tools.misc.context_provider_abstract import ContextProviderError


class WorkingMemoryFactoryError(PipelexError):
    pass


class WorkingMemoryError(PipelexError):
    pass


class WorkingMemoryConsistencyError(WorkingMemoryError):
    pass


class WorkingMemoryVariableError(WorkingMemoryError, ContextProviderError):
    pass


class WorkingMemoryTypeError(WorkingMemoryVariableError):
    pass


class WorkingMemoryStuffAttributeNotFoundError(WorkingMemoryVariableError):
    pass


class WorkingMemoryStuffNotFoundError(WorkingMemoryVariableError):
    def __init__(self, message: str, variable_name: str, pipe_code: str | None = None, concept_code: str | None = None):
        super().__init__(message, variable_name)
        self.pipe_code = pipe_code
        self.concept_code = concept_code
